import sys

try:
    from cStringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO

from twisted.application import service
from twisted.internet.endpoints import serverFromString
from twisted.internet.protocol import ServerFactory
from twisted.python.components import registerAdapter
from twisted.python import log
from ldaptor.inmemory import fromLDIFFile
from ldaptor.interfaces import IConnectedLDAPEntry
from ldaptor.protocols.ldap.ldapserver import LDAPServer

LDIF = b"""\
dn: dc=org
dc: org
objectClass: dcObject

dn: dc=example,dc=org
dc: example
objectClass: dcObject
objectClass: organization

dn: ou=people,dc=example,dc=org
objectClass: organizationalUnit
ou: people

dn: cn=bob,ou=people,dc=example,dc=org
cn: bob
gn: Bob
mail: bob@example.org
objectclass: top
objectclass: person
objectClass: inetOrgPerson
sn: Roberts
userPassword: secret

dn: gn=John+sn=Doe,ou=people,dc=example,dc=org
objectClass: addressbookPerson
gn: John
sn: Doe
street: Back alley
postOfficeBox: 123
postalCode: 54321
postalAddress: Backstreet
st: NY
l: New York City
c: US
userPassword: terces

dn: gn=John+sn=Smith,ou=people, dc=example,dc=org
objectClass: addressbookPerson
gn: John
sn: Smith
telephoneNumber: 555-1234
facsimileTelephoneNumber: 555-1235
description: This is a description that can span multi
 ple lines as long as the non-first lines are inden
 ted in the LDIF.
userPassword: eekretsay

"""


class Tree:
    def __init__(self):
        global LDIF
        self.f = BytesIO(LDIF)
        d = fromLDIFFile(self.f)
        d.addCallback(self.ldifRead)

    def ldifRead(self, result):
        self.f.close()
        self.db = result


class LDAPServerFactory(ServerFactory):
    protocol = LDAPServer

    def __init__(self, root):
        self.root = root

    def buildProtocol(self, addr):
        proto = self.protocol()
        proto.debug = self.debug
        proto.factory = self
        return proto


if __name__ == "__main__":
    from twisted.internet import reactor

    if len(sys.argv) == 2:
        port = int(sys.argv[1])
    else:
        port = 8080
    # First of all, to show logging info in stdout :
    log.startLogging(sys.stderr)
    # We initialize our tree
    tree = Tree()
    # When the LDAP Server protocol wants to manipulate the DIT, it invokes
    # `root = interfaces.IConnectedLDAPEntry(self.factory)` to get the root
    # of the DIT.  The factory that creates the protocol must therefore
    # be adapted to the IConnectedLDAPEntry interface.
    registerAdapter(lambda x: x.root, LDAPServerFactory, IConnectedLDAPEntry)
    factory = LDAPServerFactory(tree.db)
    factory.debug = True
    application = service.Application("ldaptor-server")
    myService = service.IServiceCollection(application)
    serverEndpointStr = "tcp:{}".format(port)
    e = serverFromString(reactor, serverEndpointStr)
    d = e.listen(factory)
    reactor.run()



"""
    Encoding / decoding utilities
"""

import warnings


def to_bytes(value):
    """
    Converts value to its bytes representation:

    * Uses value`s toWire method if it has one
    * Encodes to utf-8 if the value is a unicode string
    * Otherwise wraps value into bytes()
    """
    if hasattr(value, "toWire"):
        return value.toWire()
    if isinstance(value, int):
        return str(value).encode("utf-8")
    if isinstance(value, str):
        return value.encode("utf-8")
    return bytes(value)


def to_unicode(value):
    """
    Converts string to unicode:

    * Decodes value from utf-8 if it is a byte string
    * Otherwise just returns the same value
    """
    if isinstance(value, bytes):
        return value.decode("utf-8")
    return value


def get_strings(value):
    """
    Getting tuple of available string values
    (byte string and unicode string) for
    given value
    """
    if isinstance(value, str):
        return value, value.encode("utf-8")
    if isinstance(value, bytes):
        return value, value.decode("utf-8")
    return (value,)


class WireStrAlias:
    """
    A helper base or mixin class which adds __str__ method
    as an alias of toWire method but marks it as deprecated
    """

    def __str__(self):
        warnings.simplefilter("always", DeprecationWarning)
        warnings.warn(
            "{0}.__str__ method is deprecated and will not be used "
            "for getting bytes representation in the future "
            "releases, use {0}.toWire instead".format(self.__class__.__name__),
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)
        return self.toWire()

    def toWire(self):
        raise NotImplementedError("toWire method is not implemented")


class TextStrAlias:
    """
    A helper base or mixin class which adds __str__ method
    as an alias of getText method but marks it as deprecated
    """

    def __str__(self):
        warnings.simplefilter("always", DeprecationWarning)
        warnings.warn(
            "{0}.__str__ method is deprecated and will not be used "
            "for getting human readable representation in the future "
            "releases, use {0}.getText instead".format(self.__class__.__name__),
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)
        text = self.getText()
        return text

    def getText(self):
        raise NotImplementedError("getText method is not implemented")



"""
Test cases for ldaptor.protocols.ldap.proxy module.
"""

from twisted.trial import unittest
from twisted.internet import reactor, error
from ldaptor.protocols.ldap import proxy, ldaperrors
from ldaptor.protocols import pureldap
from ldaptor import testutil


class Proxy(unittest.TestCase):
    def createServer(self, *responses):
        return testutil.createServer(proxy.Proxy, *responses)

    def test_bind(self):
        server = self.createServer(
            [
                pureldap.LDAPBindResponse(resultCode=0),
            ]
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=4).toWire()
        )
        reactor.iterate()  # TODO
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=4
            ).toWire(),
        )

    def test_bind_sasl_no_credentials(self):
        # result code 14 is saslInprogress, with some server credentials.
        server = self.createServer(
            [
                pureldap.LDAPBindResponse(resultCode=14, serverSaslCreds="test123"),
            ]
        )

        server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(auth=("GSS-SPNEGO", None), sasl=True), id=4
            ).toWire()
        )
        reactor.iterate()  # TODO
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=14, serverSaslCreds="test123"),
                id=4,
            ).toWire(),
        )

    def test_search(self):
        server = self.createServer(
            [
                pureldap.LDAPBindResponse(resultCode=0),
            ],
            [
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com", [("a", ["b"])]
                ),
                pureldap.LDAPSearchResultEntry(
                    "cn=bar,dc=example,dc=com", [("b", ["c"])]
                ),
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
            ],
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire()
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPSearchRequest(), id=3).toWire()
        )
        reactor.iterate()  # TODO
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=2).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com", [("a", ["b"])]
                ),
                id=3,
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry(
                    "cn=bar,dc=example,dc=com", [("b", ["c"])]
                ),
                id=3,
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode), id=3
            ).toWire(),
        )

    def test_unbind_clientUnbinds(self):
        server = self.createServer(
            [
                pureldap.LDAPBindResponse(resultCode=0),
            ],
            [],
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire()
        )
        reactor.iterate()  # TODO
        client = server.client
        client.assertSent(pureldap.LDAPBindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=2
            ).toWire(),
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPUnbindRequest(), id=3).toWire()
        )
        server.connectionLost(error.ConnectionDone)
        reactor.iterate()  # TODO
        client.assertSent(pureldap.LDAPBindRequest(), pureldap.LDAPUnbindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=2
            ).toWire(),
        )

    def test_unbind_clientEOF(self):
        server = self.createServer(
            [
                pureldap.LDAPBindResponse(resultCode=0),
            ],
            [],
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire()
        )
        reactor.iterate()  # TODO
        client = server.client
        client.assertSent(pureldap.LDAPBindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=2
            ).toWire(),
        )
        server.connectionLost(error.ConnectionDone)
        reactor.iterate()  # TODO
        client.assertSent(
            pureldap.LDAPBindRequest(), "fake-unbind-by-LDAPClientTestDriver"
        )
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=2
            ).toWire(),
        )



"""
Test cases for ldaptor.attributeset
"""
from functools import total_ordering

from twisted.trial import unittest
from ldaptor import attributeset


class TestLDAPAttributeSet(unittest.TestCase):
    """
    Unit tests for LDAPAttributeSet.
    """

    def testEquality_True_Set(self):
        """
        Attributes are equal when the have the same key and value.
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        self.assertEqual(a, b)

    def testEquality_True_Set_Ordering(self):
        """
        The order of the element in the value doesn't matter for
        equality.
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = attributeset.LDAPAttributeSet("k", ["b", "d", "c"])
        self.assertEqual(a, b)

    def testEquality_True_List(self):
        """
        It can be compared with a list and in this case the key is
        ignored.
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = ["b", "c", "d"]
        self.assertEqual(a, b)

    def testEquality_True_List_Ordering(self):
        """
        For list comparison the order of the element don't matter.
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = ["b", "d", "c"]
        self.assertEqual(a, b)

    def testEquality_False_Value(self):
        """
        LDAPAttributeSet objects are not equal when they have
        different values.
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = attributeset.LDAPAttributeSet("k", ["b", "c", "e"])
        self.assertNotEqual(a, b)

    def testEquality_False_Key(self):
        """
        Equality fails if attributes have different keys.
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = attributeset.LDAPAttributeSet("l", ["b", "c", "d"])
        self.assertNotEqual(a, b)

    def testDifference(self):
        """
        Different operation will ignore the attribute's key and will
        perform the operation onlyb based on the attribute's value.
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = attributeset.LDAPAttributeSet("l", ["b", "c", "e"])

        result = a - b

        self.assertEqual({"d"}, result)

    def testAddNewValue(self):
        """
        Adding new value
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        a.add("e")

        self.assertEqual(a, {"b", "c", "d", "e"})

    def testAddExistingValue(self):
        """
        Adding existing value as a byte or unicode string
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])

        a.add(b"b")
        self.assertEqual(a, {"b", "c", "d"})

        a.add("b")
        self.assertEqual(a, {"b", "c", "d"})

    def testRemoveExistingValue(self):
        """
        Removing existing value as a byte or unicode string
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        a.remove(b"b")
        a.remove("c")

        self.assertEqual(a, {"d"})

    def testRemoveNonexistingValue(self):
        """
        Removing non-existing value
        """
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])

        self.assertRaises(KeyError, a.remove, "e")

    def testUnion(self):
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = attributeset.LDAPAttributeSet("k", ["b", "c", "e"])
        self.assertEqual(a | b, {"b", "c", "d", "e"})

    def testIntersection(self):
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = attributeset.LDAPAttributeSet("k", ["b", "c", "e"])
        self.assertEqual(a & b, {"b", "c"})

    def testSymmetricDifference(self):
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d"])
        b = attributeset.LDAPAttributeSet("k", ["b", "c", "e"])
        self.assertEqual(a ^ b, {"d", "e"})

    def testCopy(self):
        class Magic:
            def __lt__(self, other):
                return False

            def __gt__(self, other):
                return True

        m1 = Magic()
        a = attributeset.LDAPAttributeSet("k", ["b", "c", "d", m1])
        b = a.__copy__()
        self.assertEqual(a, b)
        self.assertNotIdentical(a, b)

        magicFromA = [val for val in a if isinstance(val, Magic)][0]
        magicFromB = [val for val in b if isinstance(val, Magic)][0]
        self.assertEqual(magicFromA, magicFromB)
        self.assertIdentical(magicFromA, magicFromB)

        a.update("x")
        self.assertEqual(a, {"b", "c", "d", m1, "x"})
        self.assertEqual(b, {"b", "c", "d", m1})

    def testDeepCopy(self):
        @total_ordering
        class Magic:
            def __eq__(self, other):
                return isinstance(other, self.__class__)

            def __hash__(self):
                return 42

            def __lt__(self, other):
                return False

        m1 = Magic()
        a = attributeset.LDAPAttributeSet("k", ["a", m1])
        b = a.__deepcopy__({})
        self.assertEqual(a, b)
        self.assertNotIdentical(a, b)

        magicFromA = [val for val in a if isinstance(val, Magic)][0]
        magicFromB = [val for val in b if isinstance(val, Magic)][0]
        self.assertEqual(magicFromA, magicFromB)
        self.assertNotIdentical(magicFromA, magicFromB)

        a.update("x")
        self.assertEqual(a, {"a", m1, "x"})
        self.assertEqual(b, {"a", m1})



"""Pythonic API for LDAP operations."""
import functools

from twisted.internet import defer
from twisted.python.failure import Failure
from zope.interface import implementer

from ldaptor.protocols.ldap import ldapclient, ldif, distinguishedname, ldaperrors
from ldaptor.protocols import pureldap, pureber
from ldaptor.samba import smbpassword
from ldaptor import ldapfilter, interfaces, delta, attributeset, entry
from ldaptor._encoder import to_bytes


class PasswordSetAggregateError(Exception):
    """Some of the password plugins failed"""

    def __init__(self, errors):
        Exception.__init__(self)
        self.errors = errors

    def __str__(self):
        return "{}: {}.".format(
            self.__doc__,
            "; ".join(
                [
                    "{} failed with {}".format(name, fail.getErrorMessage())
                    for name, fail in self.errors
                ]
            ),
        )

    def __repr__(self):
        return "<" + self.__class__.__name__ + " errors=" + repr(self.errors) + ">"


class PasswordSetAborted(Exception):
    """Aborted"""

    def __str__(self):
        return self.__doc__


class DNNotPresentError(Exception):
    """The requested DN cannot be found by the server."""


class ObjectInBadStateError(Exception):
    """The LDAP object in in a bad state."""


class ObjectDeletedError(ObjectInBadStateError):
    """The LDAP object has already been removed, unable to perform operations on it."""


class ObjectDirtyError(ObjectInBadStateError):
    """The LDAP object has a journal which needs to be committed or undone before this operation."""


class NoContainingNamingContext(Exception):
    """The server contains to LDAP naming context that would contain this object."""


class CannotRemoveRDNError(Exception):
    """The attribute to be removed is the RDN for the object and cannot be removed."""

    def __init__(self, key, val=None):
        Exception.__init__(self)
        self.key = key
        self.val = val

    def __str__(self):
        if self.val is None:
            r = repr(self.key)
        else:
            r = "{}={}".format(repr(self.key), repr(self.val))
        return (
            """The attribute to be removed, %s, is the RDN for the object and cannot be removed."""
            % r
        )


class MatchNotImplemented(NotImplementedError):
    """Match type not implemented"""

    def __init__(self, op):
        Exception.__init__(self)
        self.op = op

    def __str__(self):
        return "{}: {!r}".format(self.__doc__, self.op)


class JournaledLDAPAttributeSet(attributeset.LDAPAttributeSet):
    def __init__(self, ldapObject, *a, **kw):
        self.ldapObject = ldapObject
        super().__init__(*a, **kw)

    def add(self, value):
        self.ldapObject.journal(delta.Add(self.key, [value]))
        super().add(value)

    def update(self, sequence):
        self.ldapObject.journal(delta.Add(self.key, sequence))
        super().update(sequence)

    def remove(self, value):
        if value not in self:
            raise LookupError(value)
        self.ldapObject._canRemove(self.key, value)
        self.ldapObject.journal(delta.Delete(self.key, [value]))
        super().remove(value)

    def clear(self):
        self.ldapObject._canRemoveAll(self.key)
        super().clear()
        self.ldapObject.journal(delta.Delete(self.key))


@implementer(
    interfaces.ILDAPEntry,
    interfaces.IEditableLDAPEntry,
    interfaces.IConnectedLDAPEntry,
)
class LDAPEntryWithClient(entry.EditableLDAPEntry):
    _state = "invalid"
    """

    State of an LDAPEntry is one of:

    invalid - object not initialized yet

    ready - normal

    deleted - object has been deleted

    """

    def __init__(self, client, dn, attributes={}, complete=0):
        """

        Initialize the object.

        @param client: The LDAP client connection this object belongs
        to.

        @param dn: Distinguished Name of the object, as a string.

        @param attributes: Attributes of the object. A dictionary of
        attribute types to list of attribute values.

        """

        super().__init__(dn, attributes)
        self.client = client
        self.complete = complete

        self._journal = []

        self._remoteData = entry.EditableLDAPEntry(dn, attributes)
        self._state = "ready"

    def buildAttributeSet(self, key, values):
        return JournaledLDAPAttributeSet(self, key, values)

    def _canRemove(self, key, value):
        """

        Called by JournaledLDAPAttributeSet when it is about to remove a value
        of an attributeType.

        """
        self._checkState()
        for rdn in self.dn.split()[0].split():
            if rdn.attributeType == key and rdn.value == value:
                raise CannotRemoveRDNError(key, value)

    def _canRemoveAll(self, key):
        """

        Called by JournaledLDAPAttributeSet when it is about to remove all values
        of an attributeType.

        """
        self._checkState()
        assert not isinstance(self.dn, str)
        for keyval in self.dn.split()[0].split():
            if keyval.attributeType == key:
                raise CannotRemoveRDNError(key)

    def _checkState(self):
        if self._state != "ready":
            if self._state == "deleted":
                raise ObjectDeletedError
            else:
                raise ObjectInBadStateError(
                    "State is {} while expecting {}".format(
                        repr(self._state), repr("ready")
                    )
                )

    def journal(self, journalOperation):
        """

        Add a Modification into the list of modifications
        that need to be flushed to the LDAP server.

        Normal callers should not use this, they should use the
        o['foo']=['bar', 'baz'] -style API that enforces schema,
        handles errors and updates the cached data.

        """
        self._journal.append(journalOperation)

    # start ILDAPEntry
    def __getitem__(self, *a, **kw):
        self._checkState()
        return super().__getitem__(*a, **kw)

    def get(self, *a, **kw):
        self._checkState()
        return super().get(*a, **kw)

    def has_key(self, *a, **kw):
        self._checkState()
        return super().has_key(*a, **kw)

    def __contains__(self, key):
        self._checkState()
        return self.has_key(key)

    def keys(self):
        self._checkState()
        return super().keys()

    def items(self):
        self._checkState()
        return super().items()

    def toWire(self):
        a = []

        objectClasses = list(self.get("objectClass", []))
        objectClasses.sort()
        a.append(("objectClass", objectClasses))

        lst = list(self.items())
        lst.sort()
        for key, values in lst:
            if key != "objectClass":
                a.append((key, values))
        return ldif.asLDIF(self.dn.getText(), a)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.dn != other.dn:
            return False

        my = self.keys()
        my.sort()
        its = other.keys()
        its.sort()
        if my != its:
            return False
        for key in my:
            myAttr = self[key]
            itsAttr = other[key]
            if myAttr != itsAttr:
                return False
        return True

    def __ne__(self, other):
        return not self == other

    def __len__(self):
        return len(self.keys())

    def __nonzero__(self):
        return True

    def __hash__(self):
        return hash(self.toWire())

    def bind(self, password):
        r = pureldap.LDAPBindRequest(dn=self.dn.getText(), auth=password)
        d = self.client.send(r)
        d.addCallback(self._handle_bind_msg)
        return d

    def _handle_bind_msg(self, msg):
        assert isinstance(msg, pureldap.LDAPBindResponse)
        assert msg.referral is None  # TODO
        if msg.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)
        return self

    # end ILDAPEntry

    # start IEditableLDAPEntry
    def __setitem__(self, key, value):
        self._checkState()
        self._canRemoveAll(key)

        new = JournaledLDAPAttributeSet(self, key, value)
        super().__setitem__(key, new)
        self.journal(delta.Replace(key, value))

    def __delitem__(self, key):
        self._checkState()
        self._canRemoveAll(key)

        super().__delitem__(key)
        self.journal(delta.Delete(key))

    def undo(self):
        self._checkState()
        self._attributes.clear()
        for k, vs in self._remoteData.items():
            self._attributes[k] = self.buildAttributeSet(k, vs)
        self._journal = []

    def _assertMatchedDN(self, dn):
        assert dn == "" or dn == b""

    def _commit_success(self, msg):
        assert isinstance(msg, pureldap.LDAPModifyResponse)
        assert msg.referral is None  # TODO
        if msg.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)

        self._assertMatchedDN(msg.matchedDN)

        self._remoteData = entry.EditableLDAPEntry(self.dn, self)
        self._journal = []
        return self

    def commit(self):
        self._checkState()
        if not self._journal:
            return defer.succeed(self)

        op = pureldap.LDAPModifyRequest(
            object=self.dn.getText(), modification=[x.asLDAP() for x in self._journal]
        )
        d = defer.maybeDeferred(self.client.send, op)
        d.addCallback(self._commit_success)
        return d

    def _cbMoveDone(self, msg, newDN):
        assert isinstance(msg, pureldap.LDAPModifyDNResponse)
        assert msg.referral is None  # TODO
        if msg.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)

        self._assertMatchedDN(msg.matchedDN)
        self.dn = newDN
        return self

    def move(self, newDN):
        self._checkState()
        newDN = distinguishedname.DistinguishedName(newDN)

        newrdn = newDN.split()[0]
        newSuperior = distinguishedname.DistinguishedName(listOfRDNs=newDN.split()[1:])
        newDN = distinguishedname.DistinguishedName((newrdn,) + newSuperior.split())
        op = pureldap.LDAPModifyDNRequest(
            entry=self.dn.getText(),
            newrdn=newrdn.getText(),
            deleteoldrdn=1,
            newSuperior=newSuperior.getText(),
        )
        d = self.client.send(op)
        d.addCallback(self._cbMoveDone, newDN)
        return d

    def _cbDeleteDone(self, msg):
        assert isinstance(msg, pureldap.LDAPResult)
        if not isinstance(msg, pureldap.LDAPDelResponse):
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)
        assert msg.referral is None  # TODO
        if msg.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)

        self._assertMatchedDN(msg.matchedDN)
        return self

    def delete(self):
        self._checkState()

        op = pureldap.LDAPDelRequest(entry=self.dn.getText())
        d = self.client.send(op)
        d.addCallback(self._cbDeleteDone)
        self._state = "deleted"
        return d

    def _cbAddDone(self, msg, dn):
        assert isinstance(msg, pureldap.LDAPAddResponse), (
            "LDAPRequest response was not an LDAPAddResponse: %r" % msg
        )
        assert msg.referral is None  # TODO
        if msg.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)

        self._assertMatchedDN(msg.matchedDN)
        e = self.__class__(dn=dn, client=self.client)
        return e

    def addChild(self, rdn, attributes):
        self._checkState()

        a = []
        if attributes.get("objectClass", None):
            a.append(("objectClass", attributes["objectClass"]))
            del attributes["objectClass"]
        attributes = a + sorted(attributes.items())
        del a
        rdn = distinguishedname.RelativeDistinguishedName(rdn)
        dn = distinguishedname.DistinguishedName(listOfRDNs=(rdn,) + self.dn.split())

        ldapAttrs = []
        for attrType, values in attributes:
            ldapAttrType = pureldap.LDAPAttributeDescription(attrType)
            lst = []
            for value in values:
                if isinstance(value, str):
                    value = value.encode("utf-8")
                lst.append(pureldap.LDAPAttributeValue(value))
            ldapValues = pureber.BERSet(lst)
            ldapAttrs.append((ldapAttrType, ldapValues))
        op = pureldap.LDAPAddRequest(entry=dn.getText(), attributes=ldapAttrs)
        d = self.client.send(op)
        d.addCallback(self._cbAddDone, dn)
        return d

    def _cbSetPassword_ExtendedOperation(self, msg):
        assert isinstance(msg, pureldap.LDAPExtendedResponse)
        assert msg.referral is None  # TODO
        if msg.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)

        self._assertMatchedDN(msg.matchedDN)
        return self

    def setPassword_ExtendedOperation(self, newPasswd):
        """

        Set the password on this object.

        @param newPasswd: A string containing the new password.

        @return: A Deferred that will complete when the operation is
        done.

        """

        self._checkState()

        op = pureldap.LDAPPasswordModifyRequest(
            userIdentity=self.dn.getText(), newPasswd=newPasswd
        )
        d = self.client.send(op)
        d.addCallback(self._cbSetPassword_ExtendedOperation)
        return d

    _setPasswordPriority_ExtendedOperation = 0
    setPasswordMaybe_ExtendedOperation = setPassword_ExtendedOperation

    def setPassword_Samba(self, newPasswd, style=None):
        """

        Set the Samba password on this object.

        @param newPasswd: A string containing the new password.

        @param style: one of 'sambaSamAccount', 'sambaAccount' or
        None. Specifies the style of samba accounts used. None is
        default and is the same as 'sambaSamAccount'.

        @return: A Deferred that will complete when the operation is
        done.

        """

        self._checkState()

        nthash = smbpassword.nthash(newPasswd)
        lmhash = smbpassword.lmhash(newPasswd)

        if style is None:
            style = "sambaSamAccount"
        if style == "sambaSamAccount":
            self["sambaNTPassword"] = [nthash]
            self["sambaLMPassword"] = [lmhash]
        elif style == "sambaAccount":
            self["ntPassword"] = [nthash]
            self["lmPassword"] = [lmhash]
        else:
            raise RuntimeError("Unknown samba password style %r" % style)
        return self.commit()

    _setPasswordPriority_Samba = 20

    def setPasswordMaybe_Samba(self, newPasswd):
        """

        Set the Samba password on this object if it is a
        sambaSamAccount or sambaAccount.

        @param newPasswd: A string containing the new password.

        @return: A Deferred that will complete when the operation is
        done.

        """
        if not self.complete and not self.has_key("objectClass"):
            d = self.fetch("objectClass")
            d.addCallback(
                lambda dummy, self=self, newPasswd=newPasswd: self.setPasswordMaybe_Samba(
                    newPasswd
                )
            )
        else:
            objectClasses = [to_bytes(s.upper()) for s in self.get("objectClass", ())]
            if b"SAMBAACCOUNT" in objectClasses:
                d = self.setPassword_Samba(newPasswd, style="sambaAccount")
            elif b"SAMBASAMACCOUNT" in objectClasses:
                d = self.setPassword_Samba(newPasswd, style="sambaSamAccount")
            else:
                d = defer.succeed(self)
        return d

    def _cbSetPassword(self, dl, names):
        assert len(dl) == len(names)
        lst = []
        for name, (ok, x) in zip(names, dl):
            if not ok:
                lst.append((name, x))
        if lst:
            raise PasswordSetAggregateError(lst)
        return self

    def _cbSetPassword_one(self, result):
        return (True, None)

    def _ebSetPassword_one(self, fail):
        fail.trap(ldaperrors.LDAPException, DNNotPresentError)
        return (False, fail)

    def _setPasswordAll(self, results, newPasswd, prefix, names):
        if not names:
            return results
        name, names = names[0], names[1:]
        if results and not results[-1][0]:
            # failing
            fail = Failure(PasswordSetAborted())
            d = defer.succeed(results + [(None, fail)])
        else:
            fn = getattr(self, prefix + name)
            d = defer.maybeDeferred(fn, newPasswd)
            d.addCallbacks(self._cbSetPassword_one, self._ebSetPassword_one)

            def cb(result):
                (success, info) = result
                return results + [(success, info)]

            d.addCallback(cb)

        d.addCallback(self._setPasswordAll, newPasswd, prefix, names)
        return d

    def setPassword(self, newPasswd):
        def _passwordChangerPriorityComparison(me, other):
            mePri = getattr(self, "_setPasswordPriority_" + me)
            otherPri = getattr(self, "_setPasswordPriority_" + other)
            return (mePri > otherPri) - (mePri < otherPri)

        prefix = "setPasswordMaybe_"
        names = [name[len(prefix) :] for name in dir(self) if name.startswith(prefix)]
        names.sort(key=functools.cmp_to_key(_passwordChangerPriorityComparison))

        d = defer.maybeDeferred(self._setPasswordAll, [], newPasswd, prefix, names)
        d.addCallback(self._cbSetPassword, names)
        return d

    # end IEditableLDAPEntry

    # start IConnectedLDAPEntry

    def _cbNamingContext_Entries(self, results):
        for result in results:
            for namingContext in result.get("namingContexts", ()):
                dn = distinguishedname.DistinguishedName(namingContext)
                if dn.contains(self.dn):
                    return LDAPEntry(self.client, dn)
        raise NoContainingNamingContext(self.dn.getText())

    def namingContext(self):
        o = LDAPEntry(client=self.client, dn="")
        d = o.search(
            filterText="(objectClass=*)",
            scope=pureldap.LDAP_SCOPE_baseObject,
            attributes=["namingContexts"],
        )
        d.addCallback(self._cbNamingContext_Entries)
        return d

    def _cbFetch(self, results, overWrite):
        if len(results) != 1:
            raise DNNotPresentError(self.dn.getText())
        o = results[0]

        assert not self._journal

        if not overWrite:
            for key in list(self._remoteData.keys()):
                del self._remoteData[key]
            overWrite = o.keys()
            self.complete = 1

        for k in overWrite:
            vs = o.get(k)
            if vs is not None:
                self._remoteData[k] = vs
        self.undo()
        return self

    def fetch(self, *attributes):
        self._checkState()
        if self._journal:
            raise ObjectDirtyError(
                "cannot fetch attributes of %s, it is dirty" % repr(self)
            )

        d = self.search(scope=pureldap.LDAP_SCOPE_baseObject, attributes=attributes)
        d.addCallback(self._cbFetch, overWrite=attributes)
        return d

    def _cbSearchEntry(self, callback, objectName, attributes, complete):
        attrib = {}
        for key, values in attributes:
            attrib[to_bytes(key)] = [to_bytes(x) for x in values]
        o = LDAPEntry(
            client=self.client, dn=objectName, attributes=attrib, complete=complete
        )
        callback(o)

    def _cbSearchMsg(self, msg, controls, d, callback, complete, sizeLimitIsNonFatal):
        if isinstance(msg, pureldap.LDAPSearchResultDone):
            assert msg.referral is None  # TODO
            e = ldaperrors.get(msg.resultCode, msg.errorMessage)
            if not isinstance(e, ldaperrors.Success):
                try:
                    raise e
                except ldaperrors.LDAPSizeLimitExceeded:
                    if sizeLimitIsNonFatal:
                        pass
                except Exception:
                    d.errback(Failure())
                    return True

            # search ended successfully
            self._assertMatchedDN(msg.matchedDN)
            d.callback(controls)
            return True
        elif isinstance(msg, pureldap.LDAPSearchResultEntry):
            self._cbSearchEntry(
                callback, msg.objectName, msg.attributes, complete=complete
            )
            return False
        elif isinstance(msg, pureldap.LDAPSearchResultReference):
            return False
        else:
            raise ldaperrors.LDAPProtocolError("bad search response: %r" % msg)

    def search(
        self,
        filterText=None,
        filterObject=None,
        attributes=(),
        scope=None,
        derefAliases=None,
        sizeLimit=0,
        sizeLimitIsNonFatal=False,
        timeLimit=0,
        typesOnly=0,
        callback=None,
        controls=None,
        return_controls=False,
    ):
        self._checkState()
        d = defer.Deferred()
        if filterObject is None and filterText is None:
            filterObject = pureldap.LDAPFilterMatchAll
        elif filterObject is None and filterText is not None:
            filterObject = ldapfilter.parseFilter(filterText)
        elif filterObject is not None and filterText is None:
            pass
        elif filterObject is not None and filterText is not None:
            f = ldapfilter.parseFilter(filterText)
            filterObject = pureldap.LDAPFilter_and((f, filterObject))

        if scope is None:
            scope = pureldap.LDAP_SCOPE_wholeSubtree
        if derefAliases is None:
            derefAliases = pureldap.LDAP_DEREF_neverDerefAliases

        if attributes is None:
            attributes = ["1.1"]

        results = []
        if callback is None:
            cb = results.append
        else:
            cb = callback
        try:
            op = pureldap.LDAPSearchRequest(
                baseObject=self.dn.getText(),
                scope=scope,
                derefAliases=derefAliases,
                sizeLimit=sizeLimit,
                timeLimit=timeLimit,
                typesOnly=typesOnly,
                filter=filterObject,
                attributes=attributes,
            )
            dsend = self.client.send_multiResponse_ex(
                op,
                controls,
                self._cbSearchMsg,
                d,
                cb,
                complete=not attributes,
                sizeLimitIsNonFatal=sizeLimitIsNonFatal,
            )
        except ldapclient.LDAPClientConnectionLostException:
            d.errback(Failure())
        else:
            if callback is None:
                if return_controls:
                    d.addCallback(lambda ctls: (results, ctls))
                else:
                    d.addCallback(lambda dummy: results)

            def rerouteerr(e):
                d.errback(e)
                # returning None will stop the error
                # from being propagated and logged.

            dsend.addErrback(rerouteerr)
        return d

    def lookup(self, dn):
        e = self.__class__(self.client, dn)
        d = e.fetch("1.1")
        return d

    # end IConnectedLDAPEntry

    def __repr__(self):
        x = {}
        for key in super().keys():
            x[key] = self[key]
        keys = list(x.keys())
        keys.sort()
        a = []
        for key in keys:
            a.append("{}: {}".format(repr(key), repr(self[key])))
        attributes = ", ".join(a)
        return "{}(dn={}, attributes={{{}}})".format(
            self.__class__.__name__, repr(self.dn), attributes
        )


# API backwards compatibility
LDAPEntry = LDAPEntryWithClient


class LDAPEntryWithAutoFill(LDAPEntry):
    def __init__(self, *args, **kwargs):
        LDAPEntry.__init__(self, *args, **kwargs)
        self.autoFillers = []

    def _cb_addAutofiller(self, r, autoFiller):
        self.autoFillers.append(autoFiller)
        return r

    def addAutofiller(self, autoFiller):
        d = defer.maybeDeferred(autoFiller.start, self)
        d.addCallback(self._cb_addAutofiller, autoFiller)
        return d

    def journal(self, journalOperation):
        LDAPEntry.journal(self, journalOperation)
        for autoFiller in self.autoFillers:
            autoFiller.notify(self, journalOperation.key)



from twisted.trial import unittest
from twisted.internet import reactor, protocol, address
from ldaptor.protocols.ldap import ldapconnector, distinguishedname


class FakeProto(protocol.Protocol):
    pass


class TestCallableOverride(unittest.TestCase):
    """
    Callable values in serviceLocationOverride get to override the
    whole connecting process.
    """

    def testSimple(self):
        dn = distinguishedname.DistinguishedName("dc=example,dc=com")
        c = ldapconnector.LDAPClientCreator(reactor, FakeProto)

        def _doConnect(factory):
            factory.doStart()
            factory.startedConnecting(c)
            factory.buildProtocol(address.IPv4Address("TCP", "localhost", "1"))

        d = c.connect(
            dn,
            overrides={
                dn: _doConnect,
            },
        )

        def cb(r):
            self.failUnless(isinstance(r, FakeProto))

        d.addCallback(cb)
        return d

    def testFindOverride_plainString(self):
        """Plain strings work as override keys."""
        c = ldapconnector.LDAPConnector(
            reactor=None, dn="dc=example,dc=com", factory=None
        )
        o = c._findOverRide(
            dn=distinguishedname.DistinguishedName("cn=foo,dc=example,dc=com"),
            overrides={
                "dc=example,dc=com": ("server.example.com", 1389),
            },
        )
        self.assertEqual(o, ("server.example.com", 1389))

    def testFindOverride_root(self):
        """Empty dn can be used as override."""
        c = ldapconnector.LDAPConnector(
            reactor=None, dn="dc=example,dc=com", factory=None
        )
        o = c._findOverRide(
            dn=distinguishedname.DistinguishedName("cn=foo,dc=example,dc=com"),
            overrides={
                "": ("server.example.com", 1389),
            },
        )
        self.assertEqual(o, ("server.example.com", 1389))



"""
Test cases for ldaptor.inmemory module.
"""
from io import BytesIO

from twisted.trial import unittest

from ldaptor import inmemory, delta, testutil
from ldaptor.protocols.ldap import distinguishedname, ldaperrors


class SubclassEntry(inmemory.ReadOnlyInMemoryLDAPEntry):
    pass


class TestInMemoryDatabase(unittest.TestCase):
    def setUp(self):
        self.root = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )
        self.meta = self.root.addChild(
            rdn="ou=metasyntactic",
            attributes={
                "objectClass": ["a", "b"],
                "ou": ["metasyntactic"],
            },
        )
        self.foo = self.meta.addChild(
            rdn="cn=foo",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["foo"],
            },
        )
        self.bar = self.meta.addChild(
            rdn="cn=bar",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["bar"],
            },
        )

        self.empty = self.root.addChild(
            rdn="ou=empty",
            attributes={
                "objectClass": ["a", "b"],
                "ou": ["empty"],
            },
        )

        self.oneChild = self.root.addChild(
            rdn="ou=oneChild",
            attributes={
                "objectClass": ["a", "b"],
                "ou": ["oneChild"],
            },
        )
        self.theChild = self.oneChild.addChild(
            rdn="cn=theChild",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["theChild"],
            },
        )

    def test_children_empty(self):
        d = self.empty.children()
        d.addCallback(lambda actual: self.assertCountEqual(actual, []))
        return d

    def test_children_oneChild(self):
        d = self.oneChild.children()

        def cb(children):
            self.assertEqual(len(children), 1)
            got = [e.dn for e in children]
            want = [
                distinguishedname.DistinguishedName(
                    "cn=theChild,ou=oneChild,dc=example,dc=com"
                )
            ]
            got.sort()
            want.sort()
            self.assertCountEqual(got, want)

        d.addCallback(cb)
        return d

    def test_children_repeat(self):
        """Test that .children() returns a copy of the data so that modifying it does not affect behaviour."""
        d = self.oneChild.children()

        def cb1(children1):
            self.assertEqual(len(children1), 1)

            children1.pop()

            d = self.oneChild.children()
            return d

        d.addCallback(cb1)

        def cb2(children2):
            self.assertEqual(len(children2), 1)

        d.addCallback(cb2)
        return d

    def test_children_twoChildren(self):
        d = self.meta.children()

        def cb(children):
            self.assertEqual(len(children), 2)
            want = [
                distinguishedname.DistinguishedName(
                    "cn=foo,ou=metasyntactic,dc=example,dc=com"
                ),
                distinguishedname.DistinguishedName(
                    "cn=bar,ou=metasyntactic,dc=example,dc=com"
                ),
            ]
            got = [e.dn for e in children]
            self.assertCountEqual(got, want)

        d.addCallback(cb)
        return d

    def test_addChild(self):
        self.empty.addChild(
            rdn="a=b",
            attributes={
                "objectClass": ["a", "b"],
                "a": "b",
            },
        )
        d = self.empty.children()

        def cb(children):
            self.assertEqual(len(children), 1)
            got = [e.dn for e in children]
            want = [
                distinguishedname.DistinguishedName("a=b,ou=empty,dc=example,dc=com"),
            ]
            got.sort()
            want.sort()
            self.assertCountEqual(got, want)

        d.addCallback(cb)
        return d

    def test_addChild_Exists(self):
        self.assertRaises(
            ldaperrors.LDAPEntryAlreadyExists,
            self.meta.addChild,
            rdn="cn=foo",
            attributes={
                "objectClass": ["a"],
                "cn": "foo",
            },
        )

    def test_addChild_subclass(self):
        """
        Adding child to ReadOnlyInMemoryLDAPEntry subclass instance
        creates entry of the same class
        """
        entry = SubclassEntry(dn="dc=example,dc=com")
        child = entry.addChild(rdn="ou=empty", attributes={"objectClass": ["a"]})
        self.assertIsInstance(child, SubclassEntry)

    def test_parent(self):
        self.assertEqual(self.foo.parent(), self.meta)
        self.assertEqual(self.meta.parent(), self.root)
        self.assertEqual(self.root.parent(), None)

    def test_subtree_empty(self):
        d = self.empty.subtree()

        def cb(entries):
            self.assertEqual(len(entries), 1)

        d.addCallback(cb)
        return d

    def test_subtree_oneChild(self):
        d = self.oneChild.subtree()
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    self.oneChild,
                    self.theChild,
                ],
            )
        )
        return d

    def test_subtree_oneChild_cb(self):
        got = []
        d = self.oneChild.subtree(got.append)
        d.addCallback(self.assertEqual, None)

        def cb(dummy):
            want = [
                self.oneChild,
                self.theChild,
            ]
            self.assertCountEqual(got, want)

        d.addCallback(cb)
        return d

    def test_subtree_many(self):
        d = self.root.subtree()

        def cb(results):
            got = results
            want = [
                self.root,
                self.oneChild,
                self.theChild,
                self.empty,
                self.meta,
                self.bar,
                self.foo,
            ]
            self.assertCountEqual(got, want)

        d.addCallback(cb)
        return d

    def test_subtree_many_cb(self):
        got = []
        d = self.root.subtree(callback=got.append)

        def cb(r):
            self.assertEqual(r, None)

            want = [
                self.root,
                self.oneChild,
                self.theChild,
                self.empty,
                self.meta,
                self.bar,
                self.foo,
            ]
            self.assertCountEqual(got, want)

        d.addCallback(cb)
        return d

    def test_lookup_fail(self):
        dn = distinguishedname.DistinguishedName(
            "cn=thud,ou=metasyntactic,dc=example,dc=com"
        )
        d = self.root.lookup(dn)

        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_lookup_fail_outOfTree(self):
        dn = distinguishedname.DistinguishedName("dc=invalid")
        d = self.root.lookup(dn)

        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_lookup_deep_dn(self):
        """Entry lookup with a DN instance returns entry instance with this DN"""
        dn = distinguishedname.DistinguishedName(
            "cn=bar,ou=metasyntactic,dc=example,dc=com"
        )
        d = self.root.lookup(dn)
        d.addCallback(self.assertEqual, self.bar)
        return d

    def test_lookup_deep_str(self):
        """Entry lookup with a DN as a string returns entry instance with this DN"""
        d = self.root.lookup("cn=bar,ou=metasyntactic,dc=example,dc=com")
        d.addCallback(self.assertEqual, self.bar)
        return d

    def test_delete_root(self):
        newRoot = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )
        d = newRoot.delete()

        def eb(fail):
            fail.trap(inmemory.LDAPCannotRemoveRootError)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_delete_nonLeaf(self):
        d = self.meta.delete()

        def eb(fail):
            fail.trap(ldaperrors.LDAPNotAllowedOnNonLeaf)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_delete(self):
        d = self.foo.delete()
        d.addCallback(self.assertEqual, self.foo)
        d.addCallback(lambda _: self.meta.children())
        d.addCallback(lambda actual: self.assertCountEqual(actual, [self.bar]))
        return d

    def test_deleteChild(self):
        d = self.meta.deleteChild("cn=bar")
        d.addCallback(self.assertEqual, self.bar)
        d.addCallback(lambda _: self.meta.children())
        d.addCallback(lambda actual: self.assertCountEqual(actual, [self.foo]))
        return d

    def test_deleteChild_NonExisting(self):
        d = self.root.deleteChild("cn=not-exist")

        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_setPassword(self):
        self.foo.setPassword(b"s3krit", salt=b"\xf2\x4a")
        self.assertEqual(
            self.foo["userPassword"], [b"{SSHA}0n/Iw1NhUOKyaI9gm9v5YsO3ZInySg=="]
        )

    def test_setPassword_noSalt(self):
        self.foo.setPassword(b"s3krit")
        self.failUnless("userPassword" in self.foo)
        d = self.foo.bind(b"s3krit")
        d.addCallback(self.assertIdentical, self.foo)
        d.addCallback(lambda _: self.foo.bind(b"s4krit"))

        def eb(fail):
            fail.trap(ldaperrors.LDAPInvalidCredentials)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testSearch_withCallback(self):
        got = []
        d = self.root.search(filterText="(|(cn=foo)(cn=bar))", callback=got.append)

        def cb(r):
            self.assertEqual(r, None)

            want = [
                self.bar,
                self.foo,
            ]
            self.assertCountEqual(got, want)

        d.addCallback(cb)
        return d

    def testSearch_withoutCallback(self):
        d = self.root.search(filterText="(|(cn=foo)(cn=bar))")
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    self.bar,
                    self.foo,
                ],
            )
        )
        return d

    def test_move_noChildren_sameSuperior(self):
        d = self.empty.move("ou=moved,dc=example,dc=com")

        def getChildren(dummy):
            return self.root.children()

        d.addCallback(getChildren)
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    self.meta,
                    inmemory.ReadOnlyInMemoryLDAPEntry(
                        dn="ou=moved,dc=example,dc=com",
                        attributes={
                            "objectClass": ["a", "b"],
                            "ou": ["moved"],
                        },
                    ),
                    self.oneChild,
                ],
            )
        )
        return d

    def test_move_children_sameSuperior(self):
        d = self.meta.move("ou=moved,dc=example,dc=com")

        def getChildren(dummy):
            return self.root.children()

        d.addCallback(getChildren)
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    inmemory.ReadOnlyInMemoryLDAPEntry(
                        dn="ou=moved,dc=example,dc=com",
                        attributes={
                            "objectClass": ["a", "b"],
                            "ou": ["moved"],
                        },
                    ),
                    self.empty,
                    self.oneChild,
                ],
            )
        )
        return d

    def test_move_noChildren_newSuperior(self):
        d = self.empty.move("ou=moved,ou=oneChild,dc=example,dc=com")

        def getChildren(dummy):
            return self.root.children()

        d.addCallback(getChildren)
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    self.meta,
                    self.oneChild,
                ],
            )
        )

        def getChildren2(dummy):
            return self.oneChild.children()

        d.addCallback(getChildren2)
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    self.theChild,
                    inmemory.ReadOnlyInMemoryLDAPEntry(
                        dn="ou=moved,ou=oneChild,dc=example,dc=com",
                        attributes={
                            "objectClass": ["a", "b"],
                            "ou": ["moved"],
                        },
                    ),
                ],
            )
        )
        return d

    def test_move_children_newSuperior(self):
        d = self.meta.move("ou=moved,ou=oneChild,dc=example,dc=com")

        def getChildren(dummy):
            return self.root.children()

        d.addCallback(getChildren)
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    self.empty,
                    self.oneChild,
                ],
            )
        )

        def getChildren2(dummy):
            return self.oneChild.children()

        d.addCallback(getChildren2)
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    self.theChild,
                    inmemory.ReadOnlyInMemoryLDAPEntry(
                        dn="ou=moved,ou=oneChild,dc=example,dc=com",
                        attributes={
                            "objectClass": ["a", "b"],
                            "ou": ["moved"],
                        },
                    ),
                ],
            )
        )
        return d

    def test_commit(self):
        """ReadOnlyInMemoryLDAPEntry.commit() succeeds immediately."""
        self.meta["foo"] = ["bar"]
        d = self.meta.commit()
        self.failUnless(d.called)


class FromLDIF(unittest.TestCase):
    def test_single(self):
        ldif = BytesIO(
            b"""\
dn: cn=foo,dc=example,dc=com
objectClass: a
objectClass: b
aValue: a
aValue: b
bValue: c

"""
        )
        d = inmemory.fromLDIFFile(ldif)

        def cb1(db):
            self.assertEqual(
                db.dn, distinguishedname.DistinguishedName("cn=foo,dc=example,dc=com")
            )
            return db.children()

        d.addCallback(cb1)
        d.addCallback(self.assertEqual, [])
        return d

    def test_two(self):
        ldif = BytesIO(
            b"""\
dn: dc=example,dc=com
objectClass: dcObject
dc: example

dn: cn=foo,dc=example,dc=com
objectClass: a
cn: foo

"""
        )
        d = inmemory.fromLDIFFile(ldif)

        def cb1(db):
            self.assertEqual(
                db.dn, distinguishedname.DistinguishedName("dc=example,dc=com")
            )
            return db.subtree()

        d.addCallback(cb1)

        def cb2(children):
            self.assertEqual(len(children), 2)
            want = [
                distinguishedname.DistinguishedName("dc=example,dc=com"),
                distinguishedname.DistinguishedName("cn=foo,dc=example,dc=com"),
            ]
            got = [e.dn for e in children]
            self.assertCountEqual(got, want)

        d.addCallback(cb2)
        return d

    def test_missingNode(self):
        ldif = BytesIO(
            b"""\
dn: dc=example,dc=com
objectClass: dcObject
dc: example

dn: cn=foo,ou=nonexisting,dc=example,dc=com
objectClass: a
cn: foo

"""
        )
        d = inmemory.fromLDIFFile(ldif)

        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.failUnlessEqual(
                fail.value.toWire(), b"noSuchObject: ou=nonexisting,dc=example,dc=com"
            )

        d.addCallbacks(testutil.mustRaise, eb)
        return d


class TestDiff(unittest.TestCase):
    def testNoChange(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            "dc=example,dc=com",
            {
                "dc": ["example"],
            },
        )
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            "dc=example,dc=com",
            {
                "dc": ["example"],
            },
        )
        d = a.diffTree(b)
        d.addCallback(self.assertEqual, [])
        return d

    def testRootChange_Add(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            "dc=example,dc=com",
            {
                "dc": ["example"],
            },
        )
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            "dc=example,dc=com",
            {
                "dc": ["example"],
                "foo": ["bar"],
            },
        )
        d = a.diffTree(b)
        d.addCallback(
            self.assertEqual,
            [
                delta.ModifyOp(
                    "dc=example,dc=com",
                    [
                        delta.Add("foo", ["bar"]),
                    ],
                ),
            ],
        )
        return d

    def testChildChange_Add(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            "dc=example,dc=com",
            {
                "dc": ["example"],
            },
        )
        a.addChild(
            "cn=foo",
            {
                "cn": ["foo"],
            },
        )
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            "dc=example,dc=com",
            {
                "dc": ["example"],
            },
        )
        b.addChild(
            "cn=foo",
            {
                "cn": ["foo"],
                "foo": ["bar"],
            },
        )
        d = a.diffTree(b)
        d.addCallback(
            self.assertEqual,
            [
                delta.ModifyOp(
                    "cn=foo,dc=example,dc=com",
                    [
                        delta.Add("foo", ["bar"]),
                    ],
                ),
            ],
        )
        return d

    def testAddChild(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )

        foo = b.addChild(
            rdn="cn=foo",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["foo"],
            },
        )
        bar = b.addChild(
            rdn="cn=bar",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["bar"],
            },
        )

        d = a.diffTree(b)
        d.addCallback(
            self.assertEqual,
            [
                delta.AddOp(bar),
                delta.AddOp(foo),
            ],
        )
        return d

    def testAddSubtree(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )

        foo = b.addChild(
            rdn="ou=foo",
            attributes={
                "objectClass": ["a", "b"],
                "ou": ["foo"],
            },
        )
        baz = foo.addChild(
            rdn="cn=baz",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["baz"],
            },
        )
        bar = b.addChild(
            rdn="cn=bar",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["bar"],
            },
        )

        d = a.diffTree(b)
        d.addCallback(
            self.assertEqual,
            [
                delta.AddOp(bar),
                delta.AddOp(foo),
                delta.AddOp(baz),
            ],
        )
        return d

    def testDeleteChild(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )

        foo = a.addChild(
            rdn="cn=foo",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["foo"],
            },
        )
        bar = a.addChild(
            rdn="cn=bar",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["bar"],
            },
        )

        d = a.diffTree(b)
        d.addCallback(
            self.assertEqual,
            [
                delta.DeleteOp(bar),
                delta.DeleteOp(foo),
            ],
        )
        return d

    def testDeleteSubtree(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName("dc=example,dc=com")
        )

        foo = a.addChild(
            rdn="ou=foo",
            attributes={
                "objectClass": ["a", "b"],
                "ou": ["foo"],
            },
        )
        baz = foo.addChild(
            rdn="cn=baz",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["baz"],
            },
        )
        bar = a.addChild(
            rdn="cn=bar",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["bar"],
            },
        )

        d = a.diffTree(b)
        d.addCallback(
            self.assertEqual,
            [
                delta.DeleteOp(bar),
                delta.DeleteOp(baz),
                delta.DeleteOp(foo),
            ],
        )
        return d



"""
Test cases for ldaptor.protocols.ldap.autofill module.
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import ldapsyntax
from ldaptor.testutil import LDAPClientTestDriver


class Autofill_sum:  # TODO baseclass
    def __init__(self, resultAttr, sumAttrs):
        self.resultAttr = resultAttr
        self.sumAttrs = sumAttrs

    def start(self, ldapObject):
        pass

    def notify(self, ldapObject, attributeType):
        if attributeType not in self.sumAttrs:
            return

        sum = 0
        for sumAttr in self.sumAttrs:
            if sumAttr not in ldapObject:
                continue
            for val in ldapObject[sumAttr]:
                val = int(val)
                sum += val
        sum = str(sum)
        ldapObject[self.resultAttr] = [sum]


class LDAPAutoFill_Simple(unittest.TestCase):
    def testSimpleSum(self):
        """A simple autofiller that calculates sums of attributes should work.."""
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["some", "other"],
            },
        )
        d = o.addAutofiller(Autofill_sum(resultAttr="sum", sumAttrs=["a", "b"]))

        def cb(dummy):
            client.assertNothingSent()

            o["a"] = ["1"]
            o["b"] = ["2", "3"]

            self.failUnless("sum" in o)
            self.failUnlessEqual(o["sum"], ["6"])

        d.addCallback(cb)
        return d



"""
Test cases for ldaptor.usage
"""
import re

from twisted.python.usage import UsageError
from twisted.trial.unittest import TestCase

from ldaptor.protocols.ldap.distinguishedname import DistinguishedName
from ldaptor.usage import Options, Options_service_location, Options_scope


class ScopeOptionsImplementation(Options, Options_scope):
    """
    Minimal implementation for a command line using `Options_scope`.
    """


class TestOptions_scope(TestCase):
    def test_parseOptions_bad_scope(self):
        """
        It fails to parse the option when the scope is bad
        """
        self.assertRaisesRegex(
            UsageError,
            re.escape("bad scope: this is a bad scope"),
            ScopeOptionsImplementation().parseOptions,
            options=["--scope", "this is a bad scope"],
        )

    def test_parseOptions_default(self):
        """
        When no explicit options is provided it will set an empty dict.
        """
        sut = ServiceLocationOptionsImplementation()
        self.assertNotIn("service-location", sut.opts)

        sut.parseOptions(options=[])

        self.assertEqual({}, sut.opts["service-location"])


class ServiceLocationOptionsImplementation(Options, Options_service_location):
    """
    Minimal implementation for a command line using `Options_service_location`.
    """


class TestOptions_service_location(TestCase):
    """
    Unit tests for Options_service_location.
    """

    def test_parseOptions_default(self):
        """
        When no explicit options is provided it will set an empty dict.
        """
        sut = ServiceLocationOptionsImplementation()
        self.assertNotIn("service-location", sut.opts)

        sut.parseOptions(options=[])

        self.assertEqual({}, sut.opts["service-location"])

    def test_parseOptions_single(self):
        """
        It can have a single --service-location option.
        """
        sut = ServiceLocationOptionsImplementation()

        sut.parseOptions(
            options=["--service-location", "dc=example,dc=com:127.0.0.1:1234"]
        )

        base = DistinguishedName("dc=example,dc=com")
        value = sut.opts["service-location"][base]
        self.assertEqual(("127.0.0.1", "1234"), value)

    def test_parseOptions_invalid_DN(self):
        """
        It fails to parse the option when the base DN is not valid.
        """
        sut = ServiceLocationOptionsImplementation()

        exception = self.assertRaises(
            UsageError,
            sut.parseOptions,
            options=["--service-location", "example.com:1.2.3.4"],
        )

        self.assertEqual(
            "Invalid relative distinguished name 'example.com'.", exception.args[0]
        )

    def test_parseOptions_no_server(self):
        """
        It fails to parse the option when no host is defined, but only
        a base DN.
        """
        sut = ServiceLocationOptionsImplementation()

        exception = self.assertRaises(
            UsageError,
            sut.parseOptions,
            options=["--service-location", "dc=example,dc=com"],
        )

        self.assertEqual("service-location must specify host", exception.args[0])

    def test_parseOptions_multiple(self):
        """
        It can have have multiple --service-location options and they are
        indexed using the base DN.
        """
        sut = ServiceLocationOptionsImplementation()

        sut.parseOptions(
            options=[
                "--service-location",
                "dc=example,dc=com:127.0.0.1",
                "--service-location",
                "dc=example,dc=org:172.0.0.1",
            ]
        )

        base_com = DistinguishedName("dc=example,dc=com")
        base_org = DistinguishedName("dc=example,dc=org")
        value_com = sut.opts["service-location"][base_com]
        value_org = sut.opts["service-location"][base_org]
        self.assertEqual(("127.0.0.1", None), value_com)
        self.assertEqual(("172.0.0.1", None), value_org)



from twisted.internet import process, protocol, defer
from twisted.python import failure


class PwgenException(Exception):
    pass


class ReadPassword(protocol.ProcessProtocol):
    def __init__(self, deferred, count=1):
        self.deferred = deferred
        self.count = count
        self.stdout = ""
        self.stderr = ""

    def outReceived(self, data):
        self.stdout = self.stdout + data

    def errReceived(self, data):
        self.stderr = self.stderr + data

    def processEnded(self, reason):
        if self.stderr:
            self.deferred.errback(failure.Failure(PwgenException(reason, self.stderr)))
        elif self.stdout:
            lines = [x for x in self.stdout.split("\n") if x]
            if len(lines) != self.count:
                self.deferred.errback(
                    failure.Failure(
                        PwgenException(reason, "Wrong number of lines received.")
                    )
                )
            self.deferred.callback(lines)
        else:
            self.deferred.errback(failure.Failure(PwgenException(reason, "")))


def generate(reactor, n=1):
    assert n > 0
    d = defer.Deferred()
    proto = ReadPassword(d, n)
    process.Process(
        reactor, "pwgen", ("pwgen", "-cn1", "-N", "%d" % n), {}, None, proto
    )
    return d


if __name__ == "__main__":
    from twisted.internet import reactor
    import sys

    def say(passwords):
        for p in passwords:
            sys.stdout.write("%s\n" % p)
        return passwords

    def err(fail):
        fail.trap(PwgenException)
        sys.stderr.write("pwgen: %s\n" % fail.getErrorMessage())

    # Could get more passwords in one fork, but this stresses it more
    # on purpose.
    l = []
    for i in range(5):
        d = generate(reactor, 5)
        d.addCallbacks(say, err)
        l.append(d)

    dl = defer.DeferredList(l)
    dl.addBoth(lambda dummy: reactor.stop())
    reactor.run()



"""
Changes to the content of one single LDAP entry.

(This means these do not belong here: adding or deleting of entries,
changing of location in tree)
"""

from ldaptor import attributeset
from ldaptor.protocols import pureldap, pureber
from ldaptor.protocols.ldap import ldif, distinguishedname


class Modification(attributeset.LDAPAttributeSet):
    def patch(self, entry):
        raise NotImplementedError("%s.patch not implemented" % self.__class__.__name__)

    _LDAP_OP = None

    def asLDAP(self):
        if self._LDAP_OP is None:
            raise NotImplementedError(
                "%s.asLDAP not implemented" % self.__class__.__name__
            )
        tmplist = list(self)
        newlist = []
        for x in range(len(tmplist)):
            if isinstance(tmplist[x], str):
                value = tmplist[x].encode("utf-8")
                newlist.append(value)
            else:
                value = tmplist[x]
                newlist.append(value)

        return pureber.BERSequence(
            [
                pureber.BEREnumerated(self._LDAP_OP),
                pureber.BERSequence(
                    [
                        pureldap.LDAPAttributeDescription(self.key),
                        pureber.BERSet(map(pureldap.LDAPString, newlist)),
                    ]
                ),
            ]
        ).toWire()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return super().__eq__(other)


class Add(Modification):
    _LDAP_OP = 0

    def patch(self, entry):
        if self.key in entry:
            entry[self.key].update(self)
        else:
            entry[self.key] = self

    def asLDIF(self):
        r = []
        values = list(self)
        values.sort()
        r.append(ldif.attributeAsLDIF("add", self.key))
        for v in values:
            r.append(ldif.attributeAsLDIF(self.key, v))
        r.append(b"-\n")
        return b"".join(r)


class Delete(Modification):
    _LDAP_OP = 1

    def patch(self, entry):
        if not self:
            del entry[self.key]
        else:
            for v in self:
                entry[self.key].remove(v)

    def asLDIF(self):
        r = []
        values = list(self)
        values.sort()
        r.append(ldif.attributeAsLDIF("delete", self.key))
        for v in values:
            r.append(ldif.attributeAsLDIF(self.key, v))
        r.append(b"-\n")
        return b"".join(r)


class Replace(Modification):
    _LDAP_OP = 2

    def patch(self, entry):
        if self:
            entry[self.key] = self
        else:
            try:
                del entry[self.key]
            except KeyError:
                pass

    def asLDIF(self):
        r = []
        values = list(self)
        values.sort()
        r.append(ldif.attributeAsLDIF("replace", self.key))
        for v in values:
            r.append(ldif.attributeAsLDIF(self.key, v))
        r.append(b"-\n")
        return b"".join(r)


class Operation:
    def patch(self, root):
        """
        Find the correct entry in IConnectedLDAPEntry and patch it.

        @param root: IConnectedLDAPEntry that is at the root of the
        subtree the patch applies to.

        @returns: Deferred with None or failure.
        """
        raise NotImplementedError("%s.patch not implemented" % self.__class__.__name__)


class ModifyOp(Operation):
    def __init__(self, dn, modifications=[]):
        if not isinstance(dn, distinguishedname.DistinguishedName):
            dn = distinguishedname.DistinguishedName(stringValue=dn)
        self.dn = dn
        self.modifications = modifications[:]

    def asLDIF(self):
        r = []
        r.append(ldif.attributeAsLDIF("dn", self.dn.getText()))
        r.append(ldif.attributeAsLDIF("changetype", "modify"))
        for m in self.modifications:
            r.append(m.asLDIF())
        r.append(b"\n")
        return b"".join(r)

    def asLDAP(self):
        return pureldap.LDAPModifyRequest(
            object=self.dn.getText(),
            modification=[x.asLDAP() for x in self.modifications],
        )

    @classmethod
    def _getClassFromOp(class_, op):
        for mod in [Add, Delete, Replace]:
            if op == mod._LDAP_OP:
                return mod
        return None

    @classmethod
    def fromLDAP(class_, request):
        if not isinstance(request, pureldap.LDAPModifyRequest):
            raise RuntimeError(
                "%s.fromLDAP needs an LDAPModifyRequest" % class_.__name__
            )
        dn = request.object
        result = []
        for op, mods in request.modification:
            op = op.value
            klass = class_._getClassFromOp(op)
            if klass is None:
                raise RuntimeError(
                    "Unknown LDAP op number {!r} in {}.fromLDAP".format(
                        op, class_.__name__
                    )
                )

            key, vals = mods
            key = key.value
            vals = [x.value for x in vals]
            m = klass(key, vals)
            result.append(m)
        return class_(dn, result)

    def patch(self, root):
        d = root.lookup(self.dn)

        def gotEntry(entry, modifications):
            for mod in self.modifications:
                mod.patch(entry)
            return entry

        d.addCallback(gotEntry, self.modifications)
        return d

    def __repr__(self):
        dn = self.dn.getText()
        return (
            self.__class__.__name__
            + "("
            + "dn=%r" % dn
            + ", "
            + "modifications=%r" % self.modifications
            + ")"
        )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.dn != other.dn:
            return 0
        if self.modifications != other.modifications:
            return 0
        return 1

    def __hash__(self):
        # We use the LDIF representation as similar objects
        # should have the same LDIF.
        return hash(self.asLDIF())

    def __ne__(self, other):
        return not self == other


class AddOp(Operation):
    def __init__(self, entry):
        self.entry = entry

    def asLDIF(self):
        l = self.entry.toWire().splitlines()
        assert l[0].startswith(b"dn:")
        l[1:1] = [ldif.attributeAsLDIF("changetype", "add").rstrip(b"\n")]
        return b"".join([x + b"\n" for x in l])

    def patch(self, root):
        d = root.lookup(self.entry.dn.up())

        def gotParent(parent, entry):
            parent.addChild(entry.dn.split()[0], entry)

        d.addCallback(gotParent, self.entry)
        return d

    def __repr__(self):
        return self.__class__.__name__ + "(" + "%r" % self.entry + ")"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.entry != other.entry:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        # Use the LDIF representions as equal operations should
        # have the same LDIF.
        return hash(self.asLDIF())


class DeleteOp(Operation):
    def __init__(self, dn):
        """
        Instance can be initialized with different objects:

        * ldaptor.entry.BaseLDAPEntry instance
        * ldaptor.protocols.ldap.distinguishedname.DistinguishedName instance
        * unicode or byte string
        """
        if hasattr(dn, "dn") and isinstance(dn.dn, distinguishedname.DistinguishedName):
            self.dn = dn.dn
        elif isinstance(dn, distinguishedname.DistinguishedName):
            self.dn = dn
        elif isinstance(dn, (bytes, str)):
            self.dn = distinguishedname.DistinguishedName(stringValue=dn)
        else:
            raise AssertionError("Invalid type of object: %s" % dn.__class__.__name__)

    def asLDIF(self):
        r = []
        r.append(ldif.attributeAsLDIF("dn", self.dn.getText()))
        r.append(ldif.attributeAsLDIF("changetype", "delete"))
        r.append(b"\n")
        return b"".join(r)

    def patch(self, root):
        d = root.lookup(self.dn)

        def gotEntry(entry):
            return entry.delete()

        d.addCallback(gotEntry)
        return d

    def __repr__(self):
        dn = self.dn.getText()
        return self.__class__.__name__ + "(" + "%r" % dn + ")"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.dn != other.dn:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.dn)



"""LDAP object field value suggestion and autoupdate mechanism."""


class AutofillException(Exception):
    pass


class ObjectMissingObjectClassException(AutofillException):
    """

    The LDAPEntry is missing an objectClass this autofiller needs to
    operate.

    """



"""
Test cases for ldaptor.schema module.
"""

from twisted.trial import unittest
from ldaptor import schema
from ldaptor._encoder import to_bytes


OBJECTCLASSES = {
    "top": b"""( 2.5.6.0 NAME 'top'
    DESC 'top of the superclass chain'
    ABSTRACT
    MUST objectClass )""",
    "organization": b"""( 2.5.6.4 NAME 'organization'
    DESC 'RFC2256: an organization'
    SUP top STRUCTURAL
    MUST o
    MAY ( userPassword $ searchGuide $ seeAlso $ businessCategory $
        x121Address $ registeredAddress $ destinationIndicator $
        preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $
        telephoneNumber $ internationaliSDNNumber $
        facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $
        postalAddress $ physicalDeliveryOfficeName $ st $ l $ description ) )""",
    "organizationalUnit": b"""( 2.5.6.5 NAME 'organizationalUnit'
    DESC 'RFC2256: an organizational unit'
    SUP top STRUCTURAL
    MUST ou
    MAY ( userPassword $ searchGuide $ seeAlso $ businessCategory $
        x121Address $ registeredAddress $ destinationIndicator $
        preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $
        telephoneNumber $ internationaliSDNNumber $
        facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $
        postalAddress $ physicalDeliveryOfficeName $ st $ l $ description ) )""",
    "country": b"""( 2.5.6.2 NAME 'country'
    DESC 'RFC2256: a country'
    SUP top STRUCTURAL
    MUST c
    MAY ( searchGuide $ description ) )""",
    "dse": b"""( 1.3.6.1.4.1.4203.1.4.1 NAME ( 'OpenLDAProotDSE' 'LDAProotDSE' )
    DESC 'OpenLDAP Root DSE object'
    SUP top STRUCTURAL
    MAY cn )""",
    "person": b"""( 2.5.6.6 NAME 'person'
    DESC 'RFC2256: a person'
    SUP top STRUCTURAL
    MUST ( sn $ cn )
    MAY ( userPassword $ telephoneNumber $ seeAlso $ description ) )""",
    "obsolete": b"""( 1.3.6.1.4.1.000.1.1 NAME 'obsolete'
    DESC 'Obsolete object class'
    OBSOLETE
    SUP top STRUCTURAL )""",
    "multiple_superiors": b"""( 1.3.6.1.4.1.000.1.2 NAME 'multiple_superiors'
    DESC 'Object class with multiple superiors'
    SUP ( sup1 $ sup2 ) STRUCTURAL )""",
    "no_name": b"""( 1.3.6.1.4.1.000.1.3
    DESC 'Object class with no name'
    SUP top STRUCTURAL )""",
    "no_description": b"""( 1.3.6.1.4.1.000.1.4 NAME 'no_description'
    SUP top STRUCTURAL )""",
}


class AttributeType_KnownValues(unittest.TestCase):
    knownValues = [
        (
            b"""( 2.5.4.4 NAME ( 'sn' 'surname' )
            DESC 'RFC2256: last (family) name(s) for which the entity is known by'
            SUP name )""",
            {
                "oid": b"2.5.4.4",
                "name": (
                    b"sn",
                    b"surname",
                ),
                "desc": b"RFC2256: last (family) name(s) for which the entity is known by",
                "sup": b"name",
            },
        ),
        (
            b"""( 2.5.4.2 NAME 'knowledgeInformation'
            DESC 'RFC2256: knowledge information'
            EQUALITY caseIgnoreMatch
            SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{32768} )""",
            {
                "oid": b"2.5.4.2",
                "name": (b"knowledgeInformation",),
                "desc": b"RFC2256: knowledge information",
                "equality": b"caseIgnoreMatch",
                "syntax": b"1.3.6.1.4.1.1466.115.121.1.15{32768}",
            },
        ),
        (
            b"""( 2.5.4.5 NAME 'serialNumber'
            DESC 'RFC2256: serial number of the entity'
            EQUALITY caseIgnoreMatch
            SUBSTR caseIgnoreSubstringsMatch
            SYNTAX 1.3.6.1.4.1.1466.115.121.1.44{64} )""",
            {
                "oid": b"2.5.4.5",
                "name": (b"serialNumber",),
                "desc": b"RFC2256: serial number of the entity",
                "equality": b"caseIgnoreMatch",
                "substr": b"caseIgnoreSubstringsMatch",
                "syntax": b"1.3.6.1.4.1.1466.115.121.1.44{64}",
            },
        ),
        (
            b"""( 2.5.4.6 NAME ( 'c' 'countryName' )
            DESC 'RFC2256: ISO-3166 country 2-letter code'
            SUP name SINGLE-VALUE )""",
            {
                "oid": b"2.5.4.6",
                "name": (
                    b"c",
                    b"countryName",
                ),
                "desc": b"RFC2256: ISO-3166 country 2-letter code",
                "sup": b"name",
                "single_value": 1,
            },
        ),
        (
            b"""( 1.2.840.113549.1.9.1
            NAME ( 'email' 'emailAddress' 'pkcs9email' )
            DESC 'RFC2459: legacy attribute for email addresses in DNs'
            EQUALITY caseIgnoreIA5Match
            SUBSTR caseIgnoreIA5SubstringsMatch
            SYNTAX 1.3.6.1.4.1.1466.115.121.1.26{128} )""",
            {
                "oid": b"1.2.840.113549.1.9.1",
                "name": (
                    b"email",
                    b"emailAddress",
                    b"pkcs9email",
                ),
                "desc": b"RFC2459: legacy attribute for email addresses in DNs",
                "equality": b"caseIgnoreIA5Match",
                "substr": b"caseIgnoreIA5SubstringsMatch",
                "syntax": b"1.3.6.1.4.1.1466.115.121.1.26{128}",
            },
        ),
        (
            b"""( 1.2.840.113549.1.9.1
            NAME ( 'email' 'emailAddress' 'pkcs9email' )
            DESC 'RFC2459: legacy attribute for email addresses in DNs'
            EQUALITY caseIgnoreIA5Match
            SUBSTR caseIgnoreIA5SubstringsMatch
            SYNTAX 1.3.6.1.4.1.1466.115.121.1.26{128}
            X-ORDERED 'VALUES' )""",
            {
                "oid": b"1.2.840.113549.1.9.1",
                "name": (
                    b"email",
                    b"emailAddress",
                    b"pkcs9email",
                ),
                "desc": b"RFC2459: legacy attribute for email addresses in DNs",
                "equality": b"caseIgnoreIA5Match",
                "substr": b"caseIgnoreIA5SubstringsMatch",
                "syntax": b"1.3.6.1.4.1.1466.115.121.1.26{128}",
                "x_attrs": [
                    (b"X-ORDERED", b"VALUES"),
                ],
            },
        ),
        (
            b"""( 1.3.6.1.3.42.1 NAME 'olcDatabase'
            EQUALITY caseIgnoreMatch
            SYNTAX 1.3.6.1.4.1.1466.115.121.1.15
            SINGLE-VALUE X-ORDERED 'SIBLINGS' )""",
            {
                "oid": b"1.3.6.1.3.42.1",
                "name": (b"olcDatabase",),
                "equality": b"caseIgnoreMatch",
                "syntax": b"1.3.6.1.4.1.1466.115.121.1.15",
                "single_value": 1,
                "x_attrs": [(b"X-ORDERED", b"SIBLINGS")],
            },
        ),
        (
            b"""( 1.3.6.1.3.42.2
            NAME 'olcSuffix'
            EQUALITY distinguishedNameMatch
            SYNTAX 1.3.6.1.4.1.1466.115.121.1.12
            X-ORDERED 'VALUES' )""",
            {
                "oid": b"1.3.6.1.3.42.2",
                "name": (b"olcSuffix",),
                "equality": b"distinguishedNameMatch",
                "syntax": b"1.3.6.1.4.1.1466.115.121.1.12",
                "x_attrs": [(b"X-ORDERED", b"VALUES")],
            },
        ),
        (
            b"""( 1.3.6.1.3.42.3
            NAME 'experimentalWithQdstrings'
            X-FOO ( 'one' 'two' ) )""",
            {
                "oid": b"1.3.6.1.3.42.3",
                "name": (b"experimentalWithQdstrings",),
                "x_attrs": [(b"X-FOO", (b"one", b"two"))],
            },
        ),
        (
            b"""( 2.5.18.2 NAME 'modifyTimestamp'
            DESC 'RFC4512: time which object was last modified'
            EQUALITY generalizedTimeMatch
            ORDERING generalizedTimeOrderingMatch
            SYNTAX 1.3.6.1.4.1.1466.115.121.1.24
            SINGLE-VALUE
            NO-USER-MODIFICATION
            USAGE directoryOperation )""",
            {
                "oid": b"2.5.18.2",
                "name": (b"modifyTimestamp",),
                "desc": b"RFC4512: time which object was last modified",
                "equality": b"generalizedTimeMatch",
                "ordering": b"generalizedTimeOrderingMatch",
                "syntax": b"1.3.6.1.4.1.1466.115.121.1.24",
                "single_value": 1,
                "no_user_modification": 1,
                "usage": b"directoryOperation",
            },
        ),
        (
            b"""( 1.3.6.1.4.1.000.2.1
            NAME 'obsolete'
            OBSOLETE )""",
            {
                "oid": b"1.3.6.1.4.1.000.2.1",
                "name": (b"obsolete",),
                "obsolete": 1,
            },
        ),
        (
            b"""( 1.3.6.1.4.1.000.2.2
            NAME 'collective'
            COLLECTIVE )""",
            {
                "oid": b"1.3.6.1.4.1.000.2.2",
                "name": (b"collective",),
                "collective": 1,
            },
        ),
        (
            b"""( 1.3.6.1.4.1.000.2.3
            DESC 'Attribute type without name' )""",
            {
                "oid": b"1.3.6.1.4.1.000.2.3",
                "desc": b"Attribute type without name",
            },
        ),
    ]

    def testParse(self):
        defaults = {
            "name": None,
            "desc": None,
            "obsolete": 0,
            "sup": [],
            "equality": None,
            "ordering": None,
            "substr": None,
            "syntax": None,
            "single_value": 0,
            "collective": 0,
            "no_user_modification": 0,
            "usage": None,
        }
        for text, expected in self.knownValues:
            a = schema.AttributeTypeDescription(text)
            self.failIfEqual(a.oid, None)
            for key, want in expected.items():
                if key in defaults:
                    del defaults[key]
                got = getattr(a, key)
                self.assertEqual(got, want)

            for key, want in defaults.items():
                got = getattr(a, key)
                self.assertEqual(got, want)

    def testStringification(self):
        for want, values in self.knownValues:
            a = schema.AttributeTypeDescription(None)
            for key, val in values.items():
                setattr(a, key, val)

            want = b" ".join(want.split(None))
            got = b" ".join(to_bytes(a).split(None))
            self.assertEqual(got, want)


class ObjectClass_KnownValues(unittest.TestCase):
    knownValues = [
        (
            OBJECTCLASSES["top"],
            {
                "oid": b"2.5.6.0",
                "name": (b"top",),
                "desc": b"top of the superclass chain",
                "sup": [],
                "type": b"ABSTRACT",
                "must": [b"objectClass"],
            },
        ),
        (
            OBJECTCLASSES["organization"],
            {
                "oid": b"2.5.6.4",
                "name": (b"organization",),
                "desc": b"RFC2256: an organization",
                "sup": [b"top"],
                "type": b"STRUCTURAL",
                "must": [b"o"],
                "may": [
                    b"userPassword",
                    b"searchGuide",
                    b"seeAlso",
                    b"businessCategory",
                    b"x121Address",
                    b"registeredAddress",
                    b"destinationIndicator",
                    b"preferredDeliveryMethod",
                    b"telexNumber",
                    b"teletexTerminalIdentifier",
                    b"telephoneNumber",
                    b"internationaliSDNNumber",
                    b"facsimileTelephoneNumber",
                    b"street",
                    b"postOfficeBox",
                    b"postalCode",
                    b"postalAddress",
                    b"physicalDeliveryOfficeName",
                    b"st",
                    b"l",
                    b"description",
                ],
            },
        ),
        (
            OBJECTCLASSES["organizationalUnit"],
            {
                "oid": b"2.5.6.5",
                "name": (b"organizationalUnit",),
                "desc": b"RFC2256: an organizational unit",
                "sup": [b"top"],
                "type": b"STRUCTURAL",
                "must": [b"ou"],
                "may": [
                    b"userPassword",
                    b"searchGuide",
                    b"seeAlso",
                    b"businessCategory",
                    b"x121Address",
                    b"registeredAddress",
                    b"destinationIndicator",
                    b"preferredDeliveryMethod",
                    b"telexNumber",
                    b"teletexTerminalIdentifier",
                    b"telephoneNumber",
                    b"internationaliSDNNumber",
                    b"facsimileTelephoneNumber",
                    b"street",
                    b"postOfficeBox",
                    b"postalCode",
                    b"postalAddress",
                    b"physicalDeliveryOfficeName",
                    b"st",
                    b"l",
                    b"description",
                ],
            },
        ),
        (
            OBJECTCLASSES["dse"],
            {
                "oid": b"1.3.6.1.4.1.4203.1.4.1",
                "name": (b"OpenLDAProotDSE", b"LDAProotDSE"),
                "desc": b"OpenLDAP Root DSE object",
                "type": b"STRUCTURAL",
                "sup": [b"top"],
                "may": [b"cn"],
            },
        ),
        (
            OBJECTCLASSES["person"],
            {
                "oid": b"2.5.6.6",
                "name": (b"person",),
                "desc": b"RFC2256: a person",
                "type": b"STRUCTURAL",
                "sup": [b"top"],
                "must": [b"sn", b"cn"],
                "may": [
                    b"userPassword",
                    b"telephoneNumber",
                    b"seeAlso",
                    b"description",
                ],
            },
        ),
        (
            OBJECTCLASSES["obsolete"],
            {
                "oid": b"1.3.6.1.4.1.000.1.1",
                "name": (b"obsolete",),
                "desc": b"Obsolete object class",
                "obsolete": 1,
                "type": b"STRUCTURAL",
                "sup": [b"top"],
            },
        ),
        (
            OBJECTCLASSES["multiple_superiors"],
            {
                "oid": b"1.3.6.1.4.1.000.1.2",
                "name": (b"multiple_superiors",),
                "desc": b"Object class with multiple superiors",
                "type": b"STRUCTURAL",
                "sup": (b"sup1", b"sup2"),
            },
        ),
        (
            OBJECTCLASSES["no_name"],
            {
                "oid": b"1.3.6.1.4.1.000.1.3",
                "desc": b"Object class with no name",
                "type": b"STRUCTURAL",
                "sup": [b"top"],
            },
        ),
        (
            OBJECTCLASSES["no_description"],
            {
                "oid": b"1.3.6.1.4.1.000.1.4",
                "name": (b"no_description",),
                "type": b"STRUCTURAL",
                "sup": [b"top"],
            },
        ),
    ]

    def testParse(self):
        defaults = {
            "name": None,
            "desc": None,
            "obsolete": 0,
            "sup": None,
            "type": "STRUCTURAL",
            "must": [],
            "may": [],
        }
        for text, expected in self.knownValues:
            a = schema.ObjectClassDescription(text)
            self.failIfEqual(a.oid, None)
            for key, want in expected.items():
                if key in defaults:
                    del defaults[key]
                got = getattr(a, key)
                self.assertEqual(got, want)

            for key, want in defaults.items():
                got = getattr(a, key)
                self.assertEqual(got, want)

    def testStringification(self):
        for want, values in self.knownValues:
            a = schema.ObjectClassDescription(None)
            for key, val in values.items():
                setattr(a, key, val)

            want = b" ".join(want.split(None))
            got = b" ".join(to_bytes(a).split(None))
            self.assertEqual(got, want)


class SyntaxDescription_KnownValues(unittest.TestCase):
    knownValues = [
        (
            b"( 1.3.6.1.4.1.1466.115.121.1.3 DESC 'Attribute Type Description' )",
            {
                "oid": b"1.3.6.1.4.1.1466.115.121.1.3",
                "desc": b"Attribute Type Description",
                "human_readable": True,
                "binary_transfer_required": False,
            },
        ),
        (
            b"( 1.3.6.1.4.1.1466.115.121.1.5 DESC 'Binary' X-NOT-HUMAN-READABLE 'TRUE' )",
            {
                "oid": b"1.3.6.1.4.1.1466.115.121.1.5",
                "desc": b"Binary",
                "human_readable": False,
                "binary_transfer_required": False,
            },
        ),
        (
            b"( 1.3.6.1.4.1.1466.115.121.1.8 DESC 'Certificate' "
            b"X-BINARY-TRANSFER-REQUIRED 'TRUE' X-NOT-HUMAN-READABLE 'TRUE' )",
            {
                "oid": b"1.3.6.1.4.1.1466.115.121.1.8",
                "desc": b"Certificate",
                "human_readable": False,
                "binary_transfer_required": True,
            },
        ),
        (
            b"( 1.3.6.1.4.1.000.4.1 )",
            {
                "oid": b"1.3.6.1.4.1.000.4.1",
                "desc": None,
                "human_readable": True,
                "binary_transfer_required": False,
            },
        ),
    ]

    def testParse(self):
        for text, expected in self.knownValues:
            a = schema.SyntaxDescription(text)
            self.failIfEqual(a.oid, None)
            for key, want in expected.items():
                got = getattr(a, key)
                self.assertEqual(got, want)

    def testStringification(self):
        for want, values in self.knownValues:
            a = schema.SyntaxDescription(None)
            for key, val in values.items():
                setattr(a, key, val)

            want = b" ".join(want.split())
            got = b" ".join(to_bytes(a).split())
            self.assertEqual(got, want)


class MatchingRuleDescription_KnownValues(unittest.TestCase):
    knownValues = [
        (
            b"( 2.5.13.16 NAME 'bitStringMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.6 )",
            {
                "oid": b"2.5.13.16",
                "name": (b"bitStringMatch",),
                "desc": None,
                "obsolete": 0,
                "syntax": b"1.3.6.1.4.1.1466.115.121.1.6",
            },
        ),
        (
            b"( 1.3.6.1.4.1.000.3.1 NAME ( 'name1' 'name2' ) SYNTAX 1.3.6.1.4.1.000.4.1 )",
            {
                "oid": b"1.3.6.1.4.1.000.3.1",
                "name": (b"name1", b"name2"),
                "desc": None,
                "obsolete": 0,
                "syntax": b"1.3.6.1.4.1.000.4.1",
            },
        ),
        (
            b"( 1.3.6.1.4.1.000.3.2 NAME 'with_description' "
            b"DESC 'Matching rule with description' SYNTAX 1.3.6.1.4.1.000.4.2 )",
            {
                "oid": b"1.3.6.1.4.1.000.3.2",
                "name": (b"with_description",),
                "desc": b"Matching rule with description",
                "obsolete": 0,
                "syntax": b"1.3.6.1.4.1.000.4.2",
            },
        ),
        (
            b"( 1.3.6.1.4.1.000.3.3 NAME 'obsolete' OBSOLETE SYNTAX 1.3.6.1.4.1.000.4.3 )",
            {
                "oid": b"1.3.6.1.4.1.000.3.3",
                "name": (b"obsolete",),
                "desc": None,
                "obsolete": 1,
                "syntax": b"1.3.6.1.4.1.000.4.3",
            },
        ),
        (
            b"( 1.3.6.1.4.1.000.3.4 DESC 'Matching rule without name' SYNTAX 1.3.6.1.4.1.000.4.4 )",
            {
                "oid": b"1.3.6.1.4.1.000.3.4",
                "name": None,
                "desc": b"Matching rule without name",
                "obsolete": 0,
                "syntax": b"1.3.6.1.4.1.000.4.4",
            },
        ),
    ]

    def testParse(self):
        for text, expected in self.knownValues:
            a = schema.MatchingRuleDescription(text)
            self.failIfEqual(a.oid, None)
            for key, want in expected.items():
                got = getattr(a, key)
                self.assertEqual(got, want)

    def testStringification(self):
        for want, values in self.knownValues:
            a = schema.MatchingRuleDescription(None)
            for key, val in values.items():
                setattr(a, key, val)

            want = b" ".join(want.split())
            got = b" ".join(to_bytes(a).split())
            self.assertEqual(got, want)


class TestComparison(unittest.TestCase):
    ORDER = [
        "no_name",
        "country",
        "organization",
        "organizationalUnit",
    ]

    def setUp(self):
        data = {}
        for oc, text in OBJECTCLASSES.items():
            data[oc] = schema.ObjectClassDescription(text)
        self.data = data

    def test_eq(self):
        for k1 in self.data:
            for k2 in self.data:
                if k1 == k2:
                    self.failUnless(self.data[k1] == self.data[k2])
                else:
                    self.failIf(self.data[k1] == self.data[k2])

    def test_invalid_eq(self):
        """Object class object can be compared only to the same class object"""
        obj = schema.ObjectClassDescription(OBJECTCLASSES["top"])
        for method in (obj.__eq__, obj.__lt__, obj.__gt__):
            self.assertRaises(NotImplementedError, method, b"")

    def test_ne(self):
        for k1 in self.data:
            for k2 in self.data:
                if k1 == k2:
                    self.failIf(self.data[k1] != self.data[k2])
                else:
                    self.failUnless(self.data[k1] != self.data[k2])

    def test_order(self):
        for i, base in enumerate(self.ORDER):
            self.failUnless(self.data[base] <= self.data[base])
            self.failUnless(self.data[base] >= self.data[base])
            self.failIf(self.data[base] < self.data[base])
            self.failIf(self.data[base] > self.data[base])
            for lower in self.ORDER[:i]:
                self.failUnless(self.data[lower] < self.data[base])
                self.failUnless(self.data[lower] <= self.data[base])
                self.failIf(self.data[base] < self.data[lower])
                self.failIf(self.data[base] <= self.data[lower])
            for higher in self.ORDER[i + 1 :]:
                self.failUnless(self.data[higher] > self.data[base])
                self.failUnless(self.data[higher] >= self.data[base])
                self.failIf(self.data[base] > self.data[higher])
                self.failIf(self.data[base] >= self.data[higher])


class TestDefaultObjectClass(unittest.TestCase):
    """Structural object class type is the default one"""

    a = b"""( 1.3.6.1.4.1.003.1.1 NAME 'no_type'
    DESC 'Object class with no type' )"""
    b = b"""( 1.3.6.1.4.1.003.1.1 NAME 'no_type'
    DESC 'Object class with no type'
    STRUCTURAL )"""

    def test_default(self):
        a = schema.ObjectClassDescription(self.a)
        b = schema.ObjectClassDescription(self.b)
        self.assertEqual(a, b)


class TestInvalidObjectClass(unittest.TestCase):
    """Invalid object class definitions"""

    def test_invalid_name(self):
        text = b"( 1.1.1 NAME invalid )"
        self.assertRaises(AssertionError, schema.ObjectClassDescription, text)

    def test_invalid_multiple_name(self):
        text = b"( 1.1.1 NAME () )"
        self.assertRaises(AssertionError, schema.ObjectClassDescription, text)

    def test_empty(self):
        text = b"()"
        self.assertRaises(AssertionError, schema.ObjectClassDescription, text)


class TestInvalidAttributeType(unittest.TestCase):
    """Invalid attribute type definitions"""

    def test_invalid_name(self):
        text = b"( 1.1.1 NAME invalid )"
        self.assertRaises(AssertionError, schema.AttributeTypeDescription, text)

    def test_invalid_x_attribute(self):
        text = b"( 1.1.1 X-INVALID invalid )"
        self.assertRaises(AssertionError, schema.AttributeTypeDescription, text)

    def test_unknown_attribute(self):
        text = b"( 1.1.1 UNKNOWN 'unknown' )"
        self.assertRaises(AssertionError, schema.AttributeTypeDescription, text)


class TestInvalidMatchingRuleDescription(unittest.TestCase):
    """Invalid matching rule description definition"""

    def test_invalid_name(self):
        text = b"( 1.1.1 NAME invalid )"
        self.assertRaises(AssertionError, schema.MatchingRuleDescription, text)

    def test_no_syntax(self):
        text = b"( 1.1.1 NAME 'no_syntax' )"
        self.assertRaises(AssertionError, schema.MatchingRuleDescription, text)


"""


attributetype ( 2.5.4.7 NAME ( 'l' 'localityName' )
	DESC 'RFC2256: locality which this object resides in'
	SUP name )

attributetype ( 2.5.4.8 NAME ( 'st' 'stateOrProvinceName' )
	DESC 'RFC2256: state or province which this object resides in'
	SUP name )

attributetype ( 2.5.4.9 NAME ( 'street' 'streetAddress' )
	DESC 'RFC2256: street address of this object'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{128} )

attributetype ( 2.5.4.10 NAME ( 'o' 'organizationName' )
	DESC 'RFC2256: organization this object belongs to'
	SUP name )

attributetype ( 2.5.4.11 NAME ( 'ou' 'organizationalUnitName' )
	DESC 'RFC2256: organizational unit this object belongs to'
	SUP name )

attributetype ( 2.5.4.12 NAME 'title'
	DESC 'RFC2256: title associated with the entity'
	SUP name )

attributetype ( 2.5.4.13 NAME 'description'
	DESC 'RFC2256: descriptive information'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{1024} )

attributetype ( 2.5.4.14 NAME 'searchGuide'
	DESC 'RFC2256: search guide, obsoleted by enhancedSearchGuide'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.25 )

attributetype ( 2.5.4.15 NAME 'businessCategory'
	DESC 'RFC2256: business category'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{128} )

attributetype ( 2.5.4.16 NAME 'postalAddress'
	DESC 'RFC2256: postal address'
	EQUALITY caseIgnoreListMatch
	SUBSTR caseIgnoreListSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )

attributetype ( 2.5.4.17 NAME 'postalCode'
	DESC 'RFC2256: postal code'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{40} )

attributetype ( 2.5.4.18 NAME 'postOfficeBox'
	DESC 'RFC2256: Post Office Box'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{40} )

attributetype ( 2.5.4.19 NAME 'physicalDeliveryOfficeName'
	DESC 'RFC2256: Physical Delivery Office Name'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{128} )

attributetype ( 2.5.4.20 NAME 'telephoneNumber'
	DESC 'RFC2256: Telephone Number'
	EQUALITY telephoneNumberMatch
	SUBSTR telephoneNumberSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.50{32} )

attributetype ( 2.5.4.21 NAME 'telexNumber'
	DESC 'RFC2256: Telex Number'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.52 )

attributetype ( 2.5.4.22 NAME 'teletexTerminalIdentifier'
	DESC 'RFC2256: Teletex Terminal Identifier'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.51 )

attributetype ( 2.5.4.23 NAME ( 'facsimileTelephoneNumber' 'fax' )
	DESC 'RFC2256: Facsimile (Fax) Telephone Number'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.22 )

attributetype ( 2.5.4.24 NAME 'x121Address'
	DESC 'RFC2256: X.121 Address'
	EQUALITY numericStringMatch
	SUBSTR numericStringSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.36{15} )

attributetype ( 2.5.4.25 NAME 'internationaliSDNNumber'
	DESC 'RFC2256: international ISDN number'
	EQUALITY numericStringMatch
	SUBSTR numericStringSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.36{16} )

attributetype ( 2.5.4.26 NAME 'registeredAddress'
	DESC 'RFC2256: registered postal address'
	SUP postalAddress
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )

attributetype ( 2.5.4.27 NAME 'destinationIndicator'
	DESC 'RFC2256: destination indicator'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.44{128} )

attributetype ( 2.5.4.28 NAME 'preferredDeliveryMethod'
	DESC 'RFC2256: preferred delivery method'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.14
	SINGLE-VALUE )

attributetype ( 2.5.4.29 NAME 'presentationAddress'
	DESC 'RFC2256: presentation address'
	EQUALITY presentationAddressMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.43
	SINGLE-VALUE )

attributetype ( 2.5.4.30 NAME 'supportedApplicationContext'
	DESC 'RFC2256: supported application context'
	EQUALITY objectIdentifierMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 )

attributetype ( 2.5.4.31 NAME 'member'
	DESC 'RFC2256: member of a group'
	SUP distinguishedName )

attributetype ( 2.5.4.32 NAME 'owner'
	DESC 'RFC2256: owner (of the object)'
	SUP distinguishedName )

attributetype ( 2.5.4.33 NAME 'roleOccupant'
	DESC 'RFC2256: occupant of role'
	SUP distinguishedName )

attributetype ( 2.5.4.34 NAME 'seeAlso'
	DESC 'RFC2256: DN of related object'
	SUP distinguishedName )

attributetype ( 2.5.4.36 NAME 'userCertificate'
	DESC 'RFC2256: X.509 user certificate, use ;binary'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.8 )

attributetype ( 2.5.4.37 NAME 'cACertificate'
	DESC 'RFC2256: X.509 CA certificate, use ;binary'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.8 )

attributetype ( 2.5.4.38 NAME 'authorityRevocationList'
	DESC 'RFC2256: X.509 authority revocation list, use ;binary'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.9 )

attributetype ( 2.5.4.39 NAME 'certificateRevocationList'
	DESC 'RFC2256: X.509 certificate revocation list, use ;binary'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.9 )

attributetype ( 2.5.4.40 NAME 'crossCertificatePair'
	DESC 'RFC2256: X.509 cross certificate pair, use ;binary'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.10 )

attributetype ( 2.5.4.42 NAME ( 'givenName' 'gn' )
	DESC 'RFC2256: first name(s) for which the entity is known by'
	SUP name )

attributetype ( 2.5.4.43 NAME 'initials'
	DESC 'RFC2256: initials of some or all of names, but not the surname(s).'
	SUP name )

attributetype ( 2.5.4.44 NAME 'generationQualifier'
	DESC 'RFC2256: name qualifier indicating a generation'
	SUP name )

attributetype ( 2.5.4.45 NAME 'x500UniqueIdentifier'
	DESC 'RFC2256: X.500 unique identifier'
	EQUALITY bitStringMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.6 )

attributetype ( 2.5.4.46 NAME 'dnQualifier'
	DESC 'RFC2256: DN qualifier'
	EQUALITY caseIgnoreMatch
	ORDERING caseIgnoreOrderingMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.44 )

attributetype ( 2.5.4.47 NAME 'enhancedSearchGuide'
	DESC 'RFC2256: enhanced search guide'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.21 )

attributetype ( 2.5.4.48 NAME 'protocolInformation'
	DESC 'RFC2256: protocol information'
	EQUALITY protocolInformationMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.42 )

attributetype ( 2.5.4.50 NAME 'uniqueMember'
	DESC 'RFC2256: unique member of a group'
	EQUALITY uniqueMemberMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.34 )

attributetype ( 2.5.4.51 NAME 'houseIdentifier'
	DESC 'RFC2256: house identifier'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{32768} )

attributetype ( 2.5.4.52 NAME 'supportedAlgorithms'
	DESC 'RFC2256: supported algorithms'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.49 )

attributetype ( 2.5.4.53 NAME 'deltaRevocationList'
	DESC 'RFC2256: delta revocation list; use ;binary'
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.9 )

attributetype ( 2.5.4.54 NAME 'dmdName'
	DESC 'RFC2256: name of DMD'
	SUP name )




objectclass ( 2.5.6.3 NAME 'locality'
	DESC 'RFC2256: a locality'
	SUP top STRUCTURAL
	MAY ( street $ seeAlso $ searchGuide $ st $ l $ description ) )


objectclass ( 2.5.6.6 NAME 'person'
	DESC 'RFC2256: a person'
	SUP top STRUCTURAL
	MUST ( sn $ cn )
	MAY ( userPassword $ telephoneNumber $ seeAlso $ description ) )

objectclass ( 2.5.6.7 NAME 'organizationalPerson'
	DESC 'RFC2256: an organizational person'
	SUP person STRUCTURAL
	MAY ( title $ x121Address $ registeredAddress $ destinationIndicator $
		preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $
		telephoneNumber $ internationaliSDNNumber $
		facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $
		postalAddress $ physicalDeliveryOfficeName $ ou $ st $ l ) )

objectclass ( 2.5.6.8 NAME 'organizationalRole'
	DESC 'RFC2256: an organizational role'
	SUP top STRUCTURAL
	MUST cn
	MAY ( x121Address $ registeredAddress $ destinationIndicator $
		preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $
		telephoneNumber $ internationaliSDNNumber $ facsimileTelephoneNumber $
		seeAlso $ roleOccupant $ preferredDeliveryMethod $ street $
		postOfficeBox $ postalCode $ postalAddress $
		physicalDeliveryOfficeName $ ou $ st $ l $ description ) )

objectclass ( 2.5.6.9 NAME 'groupOfNames'
	DESC 'RFC2256: a group of names (DNs)'
	SUP top STRUCTURAL
	MUST ( member $ cn )
	MAY ( businessCategory $ seeAlso $ owner $ ou $ o $ description ) )

objectclass ( 2.5.6.10 NAME 'residentialPerson'
	DESC 'RFC2256: an residential person'
	SUP person STRUCTURAL
	MUST l
	MAY ( businessCategory $ x121Address $ registeredAddress $
		destinationIndicator $ preferredDeliveryMethod $ telexNumber $
		teletexTerminalIdentifier $ telephoneNumber $ internationaliSDNNumber $
		facsimileTelephoneNumber $ preferredDeliveryMethod $ street $
		postOfficeBox $ postalCode $ postalAddress $
		physicalDeliveryOfficeName $ st $ l ) )

objectclass ( 2.5.6.11 NAME 'applicationProcess'
	DESC 'RFC2256: an application process'
	SUP top STRUCTURAL
	MUST cn
	MAY ( seeAlso $ ou $ l $ description ) )

objectclass ( 2.5.6.12 NAME 'applicationEntity'
	DESC 'RFC2256: an application entity'
	SUP top STRUCTURAL
	MUST ( presentationAddress $ cn )
	MAY ( supportedApplicationContext $ seeAlso $ ou $ o $ l $
	description ) )

objectclass ( 2.5.6.13 NAME 'dSA'
	DESC 'RFC2256: a directory system agent (a server)'
	SUP applicationEntity STRUCTURAL
	MAY knowledgeInformation )

objectclass ( 2.5.6.14 NAME 'device'
	DESC 'RFC2256: a device'
	SUP top STRUCTURAL
	MUST cn
	MAY ( serialNumber $ seeAlso $ owner $ ou $ o $ l $ description ) )

objectclass ( 2.5.6.15 NAME 'strongAuthenticationUser'
	DESC 'RFC2256: a strong authentication user'
	SUP top AUXILIARY
	MUST userCertificate )

objectclass ( 2.5.6.16 NAME 'certificationAuthority'
	DESC 'RFC2256: a certificate authority'
	SUP top AUXILIARY
	MUST ( authorityRevocationList $ certificateRevocationList $
		cACertificate ) MAY crossCertificatePair )

objectclass ( 2.5.6.17 NAME 'groupOfUniqueNames'
	DESC 'RFC2256: a group of unique names (DN and Unique Identifier)'
	SUP top STRUCTURAL
	MUST ( uniqueMember $ cn )
	MAY ( businessCategory $ seeAlso $ owner $ ou $ o $ description ) )

objectclass ( 2.5.6.18 NAME 'userSecurityInformation'
	DESC 'RFC2256: a user security information'
	SUP top AUXILIARY
	MAY ( supportedAlgorithms ) )

objectclass ( 2.5.6.16.2 NAME 'certificationAuthority-V2'
	SUP certificationAuthority
	AUXILIARY MAY ( deltaRevocationList ) )

objectclass ( 2.5.6.19 NAME 'cRLDistributionPoint'
	SUP top STRUCTURAL
	MUST ( cn )
	MAY ( certificateRevocationList $ authorityRevocationList $
		deltaRevocationList ) )

objectclass ( 2.5.6.20 NAME 'dmd'
	SUP top STRUCTURAL
	MUST ( dmdName )
	MAY ( userPassword $ searchGuide $ seeAlso $ businessCategory $
		x121Address $ registeredAddress $ destinationIndicator $
		preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $
		telephoneNumber $ internationaliSDNNumber $ facsimileTelephoneNumber $
		street $ postOfficeBox $ postalCode $ postalAddress $
		physicalDeliveryOfficeName $ st $ l $ description ) )

objectclass ( 2.5.6.21 NAME 'pkiUser'
	DESC 'RFC2587: a PKI user'
	SUP top AUXILIARY
	MAY userCertificate )

objectclass ( 2.5.6.22 NAME 'pkiCA'
	DESC 'RFC2587: PKI certificate authority'
	SUP top AUXILIARY
	MAY ( authorityRevocationList $ certificateRevocationList $
		cACertificate $ crossCertificatePair ) )

objectclass ( 2.5.6.23 NAME 'deltaCRL'
	DESC 'RFC2587: PKI user'
	SUP top AUXILIARY
	MAY deltaRevocationList )

attributetype ( 1.3.6.1.4.1.250.1.57 NAME 'labeledURI'
	DESC 'RFC2079: Uniform Resource Identifier with optional label'
	EQUALITY caseExactMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )

objectclass ( 1.3.6.1.4.1.250.3.15 NAME 'labeledURIObject'
	DESC 'RFC2079: object that contains the URI attribute type'
	MAY ( labeledURI )
	SUP top AUXILIARY )

attributetype ( 0.9.2342.19200300.100.1.1
	NAME ( 'uid' 'userid' )
	DESC 'RFC1274: user identifier'
	EQUALITY caseIgnoreMatch
	SUBSTR caseIgnoreSubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )

attributetype ( 0.9.2342.19200300.100.1.3
	NAME ( 'mail' 'rfc822Mailbox' )
	DESC 'RFC1274: RFC822 Mailbox'
    EQUALITY caseIgnoreIA5Match
    SUBSTR caseIgnoreIA5SubstringsMatch
    SYNTAX 1.3.6.1.4.1.1466.115.121.1.26{256} )

objectclass ( 0.9.2342.19200300.100.4.19 NAME 'simpleSecurityObject'
	DESC 'RFC1274: simple security object'
	SUP top AUXILIARY
	MUST userPassword )

attributetype ( 0.9.2342.19200300.100.1.25
	NAME ( 'dc' 'domainComponent' )
	DESC 'RFC1274/2247: domain component'
	EQUALITY caseIgnoreIA5Match
	SUBSTR caseIgnoreIA5SubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )

objectclass ( 1.3.6.1.4.1.1466.344 NAME 'dcObject'
	DESC 'RFC2247: domain component object'
	SUP top AUXILIARY MUST dc )

objectclass ( 1.3.6.1.1.3.1 NAME 'uidObject'
	DESC 'RFC2377: uid object'
	SUP top AUXILIARY MUST uid )

attributetype ( 0.9.2342.19200300.100.1.37
	NAME 'associatedDomain'
	DESC 'RFC1274: domain associated with object'
	EQUALITY caseIgnoreIA5Match
	SUBSTR caseIgnoreIA5SubstringsMatch
	SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )

"""



from twisted.internet import error
from ldaptor import config, testutil
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols.pureldap import (
    LDAPBindResponse,
    LDAPBindRequest,
    LDAPUnbindRequest,
    LDAPSearchResultEntry,
    LDAPMessage,
    LDAPSearchRequest,
    LDAPSearchResultDone,
    LDAPAddRequest,
    LDAPAddResponse,
    LDAPDelRequest,
    LDAPDelResponse,
    LDAPModifyRequest,
    LDAPModifyResponse,
    LDAPModifyDNRequest,
    LDAPModifyDNResponse,
    LDAPExtendedRequest,
    LDAPExtendedResponse,
)
from twisted.test import proto_helpers
from twisted.trial import unittest

from ldaptor.protocols.ldap.merger import MergedLDAPServer


class MergedLDAPServerTest(unittest.TestCase):
    def createMergedServer(self, *responses):
        """
        Create an MergedLDAP server for testing. Initialize with
        len(responses) clients.
        :param responses: The responses to initialize the `LDAPClientTestDrives`.
        :type responses: args of lists of lists
        :return a deferred, fires when server finished connecting
        """

        def createClient(factory):
            factory.doStart()
            proto = factory.buildProtocol(addr=None)
            proto.connectionMade()

        clients = []
        for r in responses:
            clients.append(testutil.LDAPClientTestDriver(*r))

        conf = config.LDAPConfig(serviceLocationOverrides={"": createClient})
        server = MergedLDAPServer([conf for _ in clients], [False for _ in clients])
        self.clients = clients * 1
        server.protocol = lambda: clients.pop()
        server.transport = proto_helpers.StringTransport()
        server.connectionMade()

        d = server._whenConnected(lambda: server)
        return d

    def test_bind_both_success(self):
        d = self.createMergedServer(
            [[LDAPBindResponse(resultCode=0)]], [[LDAPBindResponse(resultCode=0)]]
        )

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPBindRequest(), id=4).toWire())

            self.assertEqual(
                server.transport.value(),
                LDAPMessage(LDAPBindResponse(resultCode=0), id=4).toWire(),
            )

        d.addCallback(test_f)

        return d

    def test_bind_one_invalid(self):
        d = self.createMergedServer(
            [
                [
                    LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    )
                ]
            ],
            [[LDAPBindResponse(resultCode=0)]],
        )

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPBindRequest(), id=4).toWire())
            self.assertEqual(
                server.transport.value(),
                LDAPMessage(LDAPBindResponse(resultCode=0), id=4).toWire(),
            )

        d.addCallback(test_f)
        return d

    def test_bind_both_invalid(self):
        d = self.createMergedServer(
            [
                [
                    LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    )
                ]
            ],
            [
                [
                    LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    )
                ]
            ],
        )

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPBindRequest(), id=4).toWire())
            self.assertEqual(
                server.transport.value(),
                LDAPMessage(
                    LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    ),
                    id=4,
                ).toWire(),
            )

        d.addCallback(test_f)
        return d

    def test_search_merged(self):
        d = self.createMergedServer(
            [
                [
                    LDAPSearchResultEntry("cn=foo,dc=example,dc=com", [("a", ["b"])]),
                    LDAPSearchResultEntry("cn=bar,dc=example,dc=com", [("b", ["c"])]),
                    LDAPSearchResultDone(ldaperrors.Success.resultCode),
                ]
            ],
            [
                [
                    LDAPSearchResultEntry("cn=foo,dc=example,dc=com", [("a", ["b"])]),
                    LDAPSearchResultEntry("cn=bar2,dc=example,dc=com", [("b", ["c"])]),
                    LDAPSearchResultDone(ldaperrors.Success.resultCode),
                ]
            ],
        )

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPSearchRequest(), id=3).toWire())
            self.assertEqual(
                server.transport.value(),
                LDAPMessage(
                    LDAPSearchResultEntry("cn=foo,dc=example,dc=com", [("a", ["b"])]),
                    id=3,
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultEntry("cn=bar2,dc=example,dc=com", [("b", ["c"])]),
                    id=3,
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultEntry("cn=foo,dc=example,dc=com", [("a", ["b"])]),
                    id=3,
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultEntry("cn=bar,dc=example,dc=com", [("b", ["c"])]),
                    id=3,
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultDone(ldaperrors.Success.resultCode), id=3
                ).toWire(),
            )

        d.addCallback(test_f)

        return d

    def test_search_one_invalid(self):
        d = self.createMergedServer(
            [
                [
                    LDAPSearchResultDone(
                        ldaperrors.LDAPInappropriateAuthentication.resultCode
                    )
                ]
            ],
            [
                [
                    LDAPSearchResultEntry("cn=foo,dc=example,dc=com", [("a", ["b"])]),
                    LDAPSearchResultEntry("cn=bar,dc=example,dc=com", [("b", ["c"])]),
                    LDAPSearchResultDone(ldaperrors.Success.resultCode),
                ]
            ],
        )

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPSearchRequest(), id=3).toWire())
            self.assertEqual(
                server.transport.value(),
                LDAPMessage(
                    LDAPSearchResultEntry("cn=foo,dc=example,dc=com", [("a", ["b"])]),
                    id=3,
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultEntry("cn=bar,dc=example,dc=com", [("b", ["c"])]),
                    id=3,
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultDone(ldaperrors.Success.resultCode), id=3
                ).toWire(),
            )

        d.addCallback(test_f)

        return d

    def test_unbind_clientUnbinds(self):
        d = self.createMergedServer([[]], [[]])

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPUnbindRequest(), id=3).toWire())
            server.connectionLost(error.ConnectionDone)
            for c in self.clients:
                c.assertSent(LDAPUnbindRequest())
            self.assertEqual(server.transport.value(), b"")

        d.addCallback(test_f)

        return d

    def test_unbind_clientEOF(self):
        """
        No connection is done when client has nothing to say.
        """
        d = self.createMergedServer([[]], [[]])

        def test_f(server):
            server.connectionLost(error.ConnectionDone)

            self.assertEqual([], server.clients, "A connection should not be done.")
            self.assertEqual(server.transport.value(), b"")

        d.addCallback(test_f)

        return d

    def test_unwilling_to_perform(self):
        d = self.createMergedServer([[]], [[]])

        def test_f(server):
            server.dataReceived(
                LDAPMessage(LDAPAddRequest(entry="", attributes=[]), id=3).toWire()
            )
            server.dataReceived(LDAPMessage(LDAPDelRequest(entry=""), id=4).toWire())
            server.dataReceived(
                LDAPMessage(
                    LDAPModifyRequest(object="", modification=[]), id=5
                ).toWire()
            )
            server.dataReceived(
                LDAPMessage(
                    LDAPModifyDNRequest(entry="", newrdn="", deleteoldrdn=0), id=6
                ).toWire()
            )
            server.dataReceived(
                LDAPMessage(LDAPExtendedRequest(requestName=""), id=7).toWire()
            )
            for c in server.clients:
                c.assertNothingSent()

            self.assertEqual(
                server.transport.value(),
                LDAPMessage(
                    LDAPAddResponse(
                        resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode
                    ),
                    id=3,
                ).toWire()
                + LDAPMessage(
                    LDAPDelResponse(
                        resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode
                    ),
                    id=4,
                ).toWire()
                + LDAPMessage(
                    LDAPModifyResponse(
                        resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode
                    ),
                    id=5,
                ).toWire()
                + LDAPMessage(
                    LDAPModifyDNResponse(
                        resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode
                    ),
                    id=6,
                ).toWire()
                + LDAPMessage(
                    LDAPExtendedResponse(
                        resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode
                    ),
                    id=7,
                ).toWire(),
            )

        d.addCallback(test_f)

        return d
